import pandas as pd
import glob
import os
import celescope
import scanpy as sc
import matplotlib
import matplotlib.pyplot as plt
import warnings
from celescope.tools import utils
from celescope.tools.step import s_common, Step
matplotlib.use('Agg')
warnings.filterwarnings("ignore")

TOOLS_DIR = os.path.dirname(celescope.tools.__file__)


class Mapping(Step):
    """
    ## Features

    - Output TSNE-plot of Assembled T/B Cells.

    ## Output
    - `06.mapping/{sample}_mapping.pdf` TSNE-plot of Assembled Cells.

    """
    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)
        
        self.match_dir = args.match_dir
        if self.match_dir != 'None':
            self.contig_file = glob.glob(f'{args.match_out}/matched_contig_annotations.csv')[0]

        self.h5ad = glob.glob(f'{self.match_dir}/06.analysis/*.h5ad')

    @utils.add_log
    def process_h5ad(self):
        contig_file = pd.read_csv(self.contig_file)
        assembled_cells = set(contig_file.barcode)
        print(self.h5ad)
        h5ad = sc.read_h5ad(self.h5ad[0])
        h5ad.obs['status'] = 'None'
        h5ad.obs.loc[h5ad.obs.index.isin(assembled_cells), "status"] = "T/BCR"
        
        sc.pl.tsne(h5ad, color='status', add_outline=True, legend_loc='on data',
                legend_fontsize=12, legend_fontoutline=2,frameon=False,
                title='Assembled Cells', palette=['C7','C3'])
        plt.savefig(f"{self.outdir}/{self.sample}_mapping.pdf", dpi=300)

    def run(self):
        if self.h5ad:
            self.process_h5ad()


def mapping(args):
    step_name = 'mapping'
    mapping_obj = Mapping(args, step_name)
    mapping_obj.run()


def get_opts_mapping(parser, sub_program):
    if sub_program:
        s_common(parser)
        parser.add_argument('--match_dir', help='scRNA-seq match directory', required=True)
        parser.add_argument('--match_out', help='assemble result from match directory', required=True)
    return parser


import os
import subprocess
import pysam
import pandas as pd
import json

from collections import defaultdict

from celescope.tools.emptydrop_cr import get_plot_elements
from celescope.tools.step import Step, s_common
from celescope.tools import utils
from celescope.flv_trust4.__init__ import CHAIN, PAIRED_CHAIN
from celescope.flv_CR.refine_assemble import Refine_assemble


__SUB_STEPS__ = ['mapping', 'cells', 'annotation']


class Assemble(Step):
    """
    ## Features

    - TCR/BCR Assemble by Cellranger.

    - Generate Mapping, Cells, V(D)J annotations metrics in html.

    ## Output
    
    - `03.assemble/{sample}` Cellranger vdj results.

    """
    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.fqs_dir = os.path.abspath(args.fqs_dir)
        self.mem = args.mem
        self.other_param = args.other_param
        self.ref_path = args.ref_path
        self.soft_path = args.soft_path

        # out files
        self.cmd_line = f'{self.outdir}/{self.sample}_cmd_line'


    @utils.add_log
    def assemble(self):
        """Cellranger vdj"""
        cmd = (
            f'{self.soft_path} vdj '
            f'--id={self.sample} '
            f'--reference={self.ref_path} '
            f'--fastqs={self.fqs_dir} '
            f'--sample={self.sample} '
            f'--localcores={self.thread} '
            f'--localmem={self.mem} '
        )

        if self.other_param:
            cmd += (" " + self.other_param)

        self.assemble.logger.info(cmd)
        with open(self.cmd_line, 'w') as f:
            f.write(cmd)
        
        cwd = os.getcwd()
        os.chdir(self.outdir)
        subprocess.check_call(cmd, shell=True)
        # change dir back to avoid can not find '03.assemble/stat.txt' error
        os.chdir(cwd)

    def run(self):
        self.assemble()
        if not self.args.not_refine:
            refine_aseemble = Refine_assemble(self.args)
            refine_aseemble.run()

def assemble(args):
    with Assemble(args) as runner:
        runner.run()

    with Mapping(args) as runner:
        runner.run()

    with Cells(args) as runner:
        runner.run()

    with Annotation(args) as runner:
        runner.run()


class cellranger_metrics(Step):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)

        if self.args.not_refine:
            self.out_name = self.sample
        else:
            self.out_name = self.sample + "_refine"

        cellranger_metrics_csv = f'{self.outdir}/{self.out_name}/outs/metrics_summary.csv'
        df = pd.read_csv(cellranger_metrics_csv, index_col=None)
        self.metrics_dict = df.T.to_dict()[0]

        self.seqtype = args.seqtype
        self.chains = CHAIN[self.seqtype]
        self.pairs = PAIRED_CHAIN[self.seqtype]

    def run(self):
        self.add_metrics()


class Mapping(cellranger_metrics):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)


    def add_metrics(self):
        name = 'Reads Mapped to Any V(D)J Gene'
        self.add_metric(
            name=name,
            value=self.metrics_dict[name],
            help_info="Fraction of reads that partially or wholly map to any germline V(D)J gene segment."
        )

        for chain in self.chains:
            name = f'Reads Mapped to {chain}'
            self.add_metric(
            name=name,
            value=self.metrics_dict[name],
            help_info=f"Fraction of reads that map partially or wholly to a germline {chain} gene segment."
        )


class Cells(cellranger_metrics):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)

        self.barcode_convert_json = f'{args.fqs_dir}/barcode_convert.json'

        with open(self.barcode_convert_json, 'r') as f:
            self.tenX_sgr = json.load(f)

        self.all_bam = f'{self.outdir}/{self.out_name}/outs/all_contig.bam'
        self.filter_contig_file = pd.read_csv(f'{self.outdir}/{self.out_name}/outs/filtered_contig_annotations.csv')
        
        # out
        self.count_file = f'{self.outdir}/count.txt'

    def add_metrics(self):
        name = 'Estimated Number of Cells'
        self.add_metric(
            name=name,
            value=self.metrics_dict[name],
            help_info="The number of barcodes estimated to be associated with cells that express targeted V(D)J transcripts."
        )

        name = 'Fraction Reads in Cells'
        self.add_metric(
            name=name,
            value=self.metrics_dict[name],
            help_info="Number of reads with cell-associated barcodes divided by the number of reads with valid barcodes."
        )

        name = 'Mean Reads per Cell'
        try:
            value = self.metrics_dict[name]
        except KeyError:
            value = self.metrics_dict['Mean Read Pairs per Cell']
        self.add_metric(
            name=name,
            value=value,
            help_info="Number of input read pairs divided by the estimated number of cells."
        )

        name = 'Mean Used Reads per Cell'
        try:
            value = self.metrics_dict[name]
        except KeyError:
            value = self.metrics_dict['Mean Used Read Pairs per Cell']
        self.add_metric(
            name=name,
            value=value,
            help_info="Mean number of read pairs used in assembly per cell-associated barcode."
        )

        for chain in self.chains:
            median_value = self.filter_contig_file[self.filter_contig_file['chain']==chain]['umis'].median()
            if median_value == median_value:
                self.add_metric(
                    name=f'Median used {chain} UMIs per Cell',
                    value=int(median_value),
                    help_info=f"Median number of UMIs assigned to a {chain} contig per cell."
                )
            else:
                self.add_metric(
                    name=f'Median used {chain} UMIs per Cell',
                    value=0,
                    help_info=f"Median number of UMIs assigned to a {chain} contig per cell."
                )

    @utils.add_log
    def Barcode_rank_plot(self):
        dic_umi = defaultdict(set)
        
        with pysam.AlignmentFile(self.all_bam) as fh:
            for read in fh:
                cb = read.get_tag('CB')
                umi = read.get_tag('UB')
                dic_umi[cb].add(umi)

        df_umi = pd.DataFrame()
        df_umi['barcode'] = list(dic_umi.keys())
        df_umi['UMI'] = [len(dic_umi[i]) for i in dic_umi]
        df_umi = df_umi.sort_values(by='UMI', ascending=False)
        cbs = set(self.filter_contig_file['barcode'])
        df_umi['mark'] = df_umi['barcode'].apply(lambda x: 'CB' if x in cbs else 'UB')
        df_umi['barcode'] = df_umi['barcode'].apply(lambda x : self.tenX_sgr[x.split('-')[0]])

        df_umi.to_csv(self.count_file, sep='\t', index=False)
        self.add_data(chart=get_plot_elements.plot_barcode_rank(self.count_file, log_uniform=True))

    def run(self):
        self.add_metrics()
        self.Barcode_rank_plot()


class Annotation(cellranger_metrics):
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)
    
    def add_metrics(self):
        name = 'Cells With Productive V-J Spanning Pair'
        self.add_metric(
            name=name,
            value=self.metrics_dict[name],
            help_info="Fraction of cell-associated barcodes with at least one productive contig for each chain of the receptor pair.A productive contig satisfies the following conditions: the contig annotations span the 5' end of the V region to the 3' end of the J region of the chain, a start codon was found in the expected part of the V sequence, an in-frame CDR3 amino acid motif was found, and no stop codons were found in the aligned V-J region."
        )

        for pair in self.pairs:
            chain1, chain2 = pair.split('_')[0], pair.split('_')[1]
            name = f'Cells With Productive V-J Spanning ({chain1}, {chain2}) Pair'
            self.add_metric(
                name=name,
                value=self.metrics_dict[name],
                help_info = f"Fraction of cell-associated barcodes with at least one productive contig for each chain of the ({chain1}, {chain2}) receptor pair. A productive contig satisfies the following conditions: the contig annotations span the 5' end of the V region to the 3' end of the J region of the chain, a start codon was found in the expected part of the V sequence, an in-frame CDR3 amino acid motif was found, and no stop codons were found in the aligned V-J region."
            )
        
        for chain in self.chains:
            name = f'Cells With {chain} Contig'
            self.add_metric(
                name=name,
                value=self.metrics_dict[name],
                help_info=f"Fraction of cell-associated barcodes with at least one {chain} contig annotated as a full or partial V(D)J gene."
            )

            name = f'Cells With CDR3-annotated {chain} Contig'
            self.add_metric(
                name=name,
                value=self.metrics_dict[name],
                help_info=f"Fraction of cell-associated barcodes with at least one {chain} contig where a CDR3 was detected."
            )

            name = f'Cells With V-J Spanning {chain} Contig'
            self.add_metric(
                name=name,
                value=self.metrics_dict[name],
                help_info=f"Fraction of cell-associated barcodes with at least one contig spanning the 5' end of the V region to the 3' end of the J region for {chain}."
            )

            name = f'Cells With Productive {chain} Contig'
            self.add_metric(
                name=name,
                value=self.metrics_dict[name],
                help_info=f"Fraction of cell-associated barcodes with at least one contig that spans the 5' end of the V region to the 3' end of the J region for {chain}, has a start codon in the expected part of the V sequence, has an in-frame CDR3, and has no stop codons in the aligned V-J region."
            )

    
def get_opts_assemble(parser, sub_program):
    parser.add_argument('--ref_path', help='reference path for cellranger')
    parser.add_argument('--soft_path', help='soft path for cellranger')
    parser.add_argument('--other_param', help='Other cellranger parameters.', default="")
    parser.add_argument('--mem', help='memory(G)', default=10)
    parser.add_argument('--seqtype', help='TCR or BCR', choices=['TCR', 'BCR'], required=True)
    parser.add_argument("--not_refine", help="Do not perform the refine step. ", action="store_true")
    parser.add_argument("--coeff",
        help="coefficient will affect auto and snr noise filter, recommend 1.5 for auto", 
        type=float, default=1.5)
    if sub_program:
        s_common(parser)
        parser.add_argument('--fqs_dir', help='fastq dir after convert', required=True)
    return parser




from celescope.tools import utils
from celescope.tools.mkref import Mkref, super_opts


class Mkref_rna(Mkref):
    """
    ## Features
    - Create a genome reference directory.

    ## Usage
    ```
    celescope utils mkgtf Homo_sapiens.GRCh38.99.gtf Homo_sapiens.GRCh38.99.filtered.gtf
    celescope rna mkref \\
    --genome_name Homo_sapiens_ensembl_99_filtered \\
    --fasta Homo_sapiens.GRCh38.dna.primary_assembly.fa \\
    --gtf Homo_sapiens.GRCh38.99.filtered.gtf
    ```

    ## Output

    - STAR genome index files

    - Genome config file
    ```
    $ cat celescope_genome.config
    [genome]
    genome_name = Homo_sapiens_ensembl_99
    genome_type = rna
    fasta = Homo_sapiens.GRCh38.dna.primary_assembly.fa
    gtf = Homo_sapiens.GRCh38.99.gtf
    ```
    """


    @utils.add_log
    def build_rna_star_index(self):
        cmd = (
            f'STAR \\\n'
            f'--runMode genomeGenerate \\\n'
            f'--runThreadN {self.thread} \\\n'
            f'--genomeDir ./ \\\n'
            f'--genomeFastaFiles {self.fasta} \\\n'
            f'--sjdbGTFfile {self.gtf} \\\n'
            f'--sjdbOverhang 100 \\\n'
        )
        if self.STAR_param:
            cmd += (" " + self.STAR_param)
        self.build_star_index.logger.info(cmd)
        self.debug_subprocess_call(cmd)


    @staticmethod
    def parse_genomeDir(genomeDir):
        return Mkref.parse_genomeDir(genomeDir, files=('gtf', 'mt_gene_list'))


    @utils.add_log
    def run(self):
        super().run()
        self.build_star_index()



def mkref(args):
    genome_type = 'rna'
    # files do not contain refflat because refflat is not input argument
    with Mkref_rna(genome_type, args, files=('gtf', 'mt_gene_list')) as runner:
        runner.run()


def get_opts_mkref(parser, sub_program):
    super_opts(parser, sub_program)
    if sub_program:
        parser.add_argument(
            "--gtf",
            help="Required. Genome gtf file. Use absolute path or relative path to `genomeDir`.",
            required=True
        )
        parser.add_argument(
            "--mt_gene_list",
            help="""Mitochondria gene list file. Use absolute path or relative path to `genomeDir`.
It is a plain text file with one gene per line. 
If not provided, will use `MT-` and `mt-` to determine mitochondria genes.""",
            default="None"
        )



#!/bin/env python
# coding=utf8

import os
import sys
import subprocess
import pandas as pd
import pysam
from celescope.tools.step import Step, s_common
from celescope.tools import utils

toolsdir = os.path.dirname(__file__)


class Replacement(Step):
    """
    Features
    - Computes the replacement rates in each cell and gene.
    - Boxplots for rates distribution.

    Output
    - `{sample}.new_matrix.tsv.gz` New RNA matrix.
    - `{sample}.old_matrix.tsv.gz` Old RNA matrix.
    - `{sample}.fraction_of_newRNA_per_cell.txt` Fraction of new RNA of each cell.
    - `{sample}.fraction_of_newRNA_per_gene.txt` Fraction of new RNA of each gene.
    - `{sample}.fraction_of_newRNA_matrix.txt` Fraction of new RNA of each cell and gene.
    """

    def __init__(self, args):
        Step.__init__(self, args)

        # input files
        self.outdir = args.outdir
        self.sample = args.sample
        self.bam_file = args.bam
        self.snp_file = args.bg
        self.bg_cov = args.bg_cov
        self.snp_threshold = args.snp_threshold

        # output files
        self.outmat = os.path.join(self.outdir, self.sample+'.TC_matrix.tsv')
        self.outpre = os.path.join(self.outdir, self.sample)

    @utils.add_log
    def run(self):
        # get backgroud snp        
        bg = self.background_snp(self.snp_file, self.bg_cov, self.snp_threshold)
        # get reads with TC
        outframe = self.extract_dem(self.bam_file, bg)
        # run_R
        self.generate_TC_matrix(outframe, self.outmat)

        # split to New and Old Matrix
        new_mat = self.outpre+'.new_matrix.tsv'
        old_mat = self.outpre+'.old_matrix.tsv'
        con_mat = self.outpre+'.NvsO_matrix.tsv'
        self.split_matrix(self.outmat, self.outpre)

        # replacement stat
        self.replacment_stat(con_mat, self.outpre)
        # plot
        div_item = self.replacment_plot(self.outpre)

        # report
        self.report_prepare(div_item)
        self._clean_up()

        # clean
        cmd = ['rm', self.outmat]
        self.run_cmd(cmd)
        cmd = ['rm', con_mat]
        self.run_cmd(cmd)
        cmd = ['gzip', new_mat]
        self.run_cmd(cmd)
        cmd = ['gzip', old_mat]
        self.run_cmd(cmd)

    def run_cmd(self, cmd):
        subprocess.call(' '.join(cmd), shell=True)

    @utils.add_log
    def extract_dem(self, bam, bg):
        bamfile = pysam.AlignmentFile(bam, 'rb')
        countdict = {}
        for read in bamfile.fetch():
            try:
                chro = read.reference_name
                cb = read.get_tag('CB')
                ub = read.get_tag('UB')
                if not read.has_tag('GN'):
                    continue
                gene = read.get_tag('GN')

                if read.get_tag('ST') == '+':
                    stag = read.get_tag('TL')
                else:
                    stag = read.get_tag('AL')
                if len(stag) == 1 and stag[0] == 0:
                    gene += '--T'
                else:
                    fcount = 0
                    for si in range(0, len(stag)):
                        pos = chro + '_' + str(stag[si])
                        if pos in bg:
                            fcount += 1
                    if fcount == len(stag):
                        gene += '--T'
                    else:
                        gene += '--C'

                readinfo = '\t'.join([gene, cb, ub])
                if readinfo not in countdict:
                    countdict[readinfo] = 1

            except (ValueError, KeyError):
                continue
        bamfile.close()

        out1 = []
        for rid in countdict:
            checkt = rid.split('--T')
            if len(checkt) == 2:
                pairc = rid.replace('--T','--C')
                if pairc not in countdict:
                    out1.append(rid.split('\t'))
            else:
                out1.append(rid.split('\t'))

        outframe = pd.DataFrame(out1)
        outframe.columns=['geneID','Barcode','UMI']
        out1=[]
        countdict = {}

        return outframe

    @utils.add_log
    def background_snp(self, bgfiles, cov=1, snp_threshold=0.5):
        ## dict.update()
        outdict = {}
        bgs=bgfiles.strip().split(',')
        for bgfile in bgs:
            if bgfile.endswith('.csv'):
                df = pd.read_csv(bgfile,index_col=0, dtype={"chrom":str})
                df = df[df['convs']>=cov]
                if 'pos' in df.columns:
                    df['chrpos'] = df['chrom']+'_'+df['pos'].astype(str)
                else: #compatible with previous version
                    df['chrpos'] = df['chrom']+'_'+df['pos2'].astype(str)
                df1 = df[['chrpos','posratio']]
                df1.set_index('chrpos',inplace=True)
                outdict.update(df1.to_dict(orient='index'))
            elif bgfile.endswith('.vcf'):
                bcf_in = pysam.VariantFile(bgfile)
                for rec in bcf_in.fetch():
                    try:
                        chrom, pos = rec.chrom, rec.pos
                        chr_pos = chrom+'_'+str(pos-1)
                        outdict[chr_pos] = 1
                    except (ValueError, KeyError):
                        continue
                bcf_in.close()
            elif bgfile.upper() == "SELF":
                selfbg = os.path.splitext(self.bam_file)[0]+'.csv'
                df = pd.read_csv(selfbg,index_col=0, dtype={"chrom":str})
                df = df[df['convs']>=cov]
                df = df[df['posratio']>=snp_threshold]
                if 'pos' in df.columns:
                    df['chrpos'] = df['chrom']+'_'+df['pos'].astype(str)
                else: #compatible with previous version
                    df['chrpos'] = df['chrom']+'_'+df['pos2'].astype(str)
                df1 = df[['chrpos','posratio']]
                df1.set_index('chrpos',inplace=True)
                outdict.update(df1.to_dict(orient='index'))                
                continue
            else:
                try:
                    sys.exit(1)
                except SystemExit:
                    print('Background snp file format cannot be recognized! Only csv or vcf format.')
                finally:
                    print('Background snp file format cannot be recognized! Only csv or vcf format.')
        return outdict

    @utils.add_log
    def generate_TC_matrix(self, read, outmat):
        table = read.pivot_table(
                    index='geneID', columns='Barcode', values='UMI',
                    aggfunc=len).fillna(0).astype(int)
        table.index.name = ''
        table.to_csv(outmat,sep="\t")


    @utils.add_log
    def split_matrix(self, mat, outpre):
        outnew = open(outpre+'.new_matrix.tsv', 'w')
        outold = open(outpre+'.old_matrix.tsv', 'w')
        con_mat = open(outpre+'.NvsO_matrix.tsv', 'w')
        infile = open(mat, 'r')

        tmph = infile.readline().strip().split()
        fill_na = ['0'] * len(tmph)
        tmph.insert(0, '')
        outnew.write('\t'.join(tmph)+'\n')
        outold.write('\t'.join(tmph)+'\n')
        con_mat.write('\t'.join(tmph)+'\n')

        genes = {}
        for i in infile:
            ii = i.strip().split()
            gt = ii[0].split('--')
            ii[0] = gt[0]
            if gt[0] not in genes:
                genes[gt[0]] = [0, [], []]
            if gt[1] == 'C':
                genes[gt[0]][0] += 1
                genes[gt[0]][1] = ii[1:]
                outnew.write('\t'.join(ii)+'\n')
            elif gt[1] == 'T':
                genes[gt[0]][0] += 2
                genes[gt[0]][2] = ii[1:]
                outold.write('\t'.join(ii)+'\n')

        for gi in genes:
            con_mat.write(gi)
            if genes[gi][0] == 3:
                for ci in range(len(genes[gi][1])):
                    con_mat.write('\t'+genes[gi][1][ci]+':'+genes[gi][2][ci])
                con_mat.write('\n')
            elif genes[gi][0] == 1:
                outold.write(gi+'\t'+'\t'.join(fill_na)+'\n')
                for ci in range(len(genes[gi][1])):
                    con_mat.write('\t'+genes[gi][1][ci]+':'+'0')
                con_mat.write('\n')
            elif genes[gi][0] == 2:
                outnew.write(gi+'\t'+'\t'.join(fill_na)+'\n')
                for ci in range(len(genes[gi][2])):
                    con_mat.write('\t'+'0'+':'+genes[gi][2][ci])
                con_mat.write('\n')

        outnew.close()
        outold.close()
        con_mat.close()
        infile.close()

    @utils.add_log
    def replacment_stat(self, inmat, outpre, mincell=10, mingene=10):

        outcell = open(outpre+'.fraction_of_newRNA_per_cell.txt', 'w')
        outgene = open(outpre+'.fraction_of_newRNA_per_gene.txt', 'w')
        outmat1 = open(outpre+'.fraction_of_newRNA_matrix.txt', 'w')

        cells = {}
        genes = {}
        mats = {}
        with open(inmat) as f:
            hh = f.readline().strip().split()
            hh.insert(0, '')
            outmat1.write('\t'.join(hh)+'\n')
            for h in hh[1:]:
                cells[h] = [[], []]
            for i in f:
                ii = i.strip().split()
                mats[ii[0]] = []
                genes[ii[0]] = [[], []]
                for xi in range(1, len(ii)):
                    xx = [int(x) for x in ii[xi].split(':')]
                    if sum(xx) == 0:
                        tmpf = 'NA'
                    else:
                        tmpf = float(xx[0])/(xx[0]+xx[1])
                    mats[ii[0]].append(str(tmpf))
                    if sum(xx) < 2:
                        continue
                    cells[hh[xi]][0].append(float(xx[0]))
                    cells[hh[xi]][1].append(xx[1])
                    genes[ii[0]][0].append(float(xx[0]))
                    genes[ii[0]][1].append(xx[1])

        for ci in cells:
            if len(cells[ci][0]) < mincell:
                continue
            cfloat = sum(cells[ci][0])/(sum(cells[ci][0])+sum(cells[ci][1]))
            outcell.write(ci+'\t'+str(cfloat)+'\n')

        for gi in genes:
            if len(genes[gi][0]) < mingene:
                continue
            gfloat = sum(genes[gi][0])/(sum(genes[gi][0])+sum(genes[gi][1]))
            outgene.write(gi+'\t'+str(gfloat)+'\n')

        for mi in mats:
            outmat1.write(mi+'\t'+'\t'.join(mats[mi])+'\n')

        outcell.close()
        outgene.close()
        outmat1.close()

    @utils.add_log
    def replacment_plot(self, sample):
        import plotly
        import plotly.graph_objects as go
        from plotly.subplots import make_subplots

        outpre = os.path.basename(sample)
        df1 = pd.read_table(sample+'.fraction_of_newRNA_per_gene.txt', header=None)
        df1.columns = ['gene', 'per']
        df = pd.read_table(sample+'.fraction_of_newRNA_per_cell.txt', header=None)
        df.columns = ['gene', 'per']

        fig = make_subplots(rows=1, cols=2)
        fig.add_trace(
            go.Violin(y=df1['per'], box_visible=True, line_color='black',
                      meanline_visible=True, fillcolor='#1f77b4', opacity=0.6, x0=outpre),
            row=1, col=1
        )
        fig.add_trace(
            go.Violin(y=df['per'], box_visible=True, line_color='black',
                      meanline_visible=True, fillcolor='#ff7f0e', opacity=0.6, x0=outpre),
            row=1, col=2
        )

        fig.update_layout(yaxis_zeroline=True,  showlegend=False)
        fig.update_layout(plot_bgcolor='#FFFFFF')
        fig.update_xaxes(showgrid=False, linecolor='black', showline=True, ticks=None)
        fig.update_yaxes(showgrid=False, linecolor='black', showline=True, ticks='outside',
                         title_text="Fraction new RNA (per gene)", row=1, col=1, rangemode="tozero")
        fig.update_yaxes(showgrid=False, linecolor='black', showline=True, ticks='outside',
                         title_text="Fraction new RNA (per cell)", row=1, col=2, rangemode="tozero")

        div = plotly.offline.plot(fig, include_plotlyjs=False, output_type='div')

        return div

    def report_prepare(self, outdiv):
        self.add_data(replacement=outdiv)


@utils.add_log
def replacement(args):

    with Replacement(args) as runner:
        runner.run()


def get_opts_replacement(parser, sub_program):
    parser.add_argument('--bg_cov', type=int, default=1,
                        help='background snp depth filter, lower than bg_cov will be discarded. Only valid in csv format')
    parser.add_argument('--snp_threshold', type=float, default=0.5,
                        help='snp threshold filter, greater than snp_threshold will be recognized as snp. Only valid in csv format')
    if sub_program:
        parser.add_argument('--bam', help='bam file from conversion step', required=True)
        parser.add_argument('--bg', help='background snp file, csv or vcf format', required=True)
        #parser.add_argument('--cell_keep', type=int, default=100000, help='filter cell')
        parser.add_argument('--min_cell', type=int, default=10, help='a gene expressed in at least cells, default 10')
        parser.add_argument('--min_gene', type=int, default=10, help='at least gene num in a cell, default 10')
        parser = s_common(parser)
    return parser



import subprocess
import os

from celescope.tools import utils
from celescope.tools.step import Step, s_common


class Cellranger(Step):
    """
    ## Features
    - Single cell RNA-seq Gene Expression analysis by Cellranger.
    ## Output
    - `03.assemble/{sample}` Cellranger count results.
    """
    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.fqs_dir = os.path.abspath(args.fqs_dir)
        self.mem = args.mem
        self.other_param = args.other_param
        self.ref_path = args.ref_path
        self.soft_path = args.soft_path

        # out files
        self.cmd_line = f'{self.outdir}/{self.sample}_cmd_line'

    @utils.add_log
    def cellranger_count(self):
        """Cellranger count"""
        cmd = (
            f'{self.soft_path} count '
            f'--id={self.sample} '
            f'--transcriptome={self.ref_path} '
            f'--fastqs={self.fqs_dir} '
            f'--sample={self.sample} '
            f'--localcores={self.thread} '
            f'--localmem={self.mem} '
        )

        if self.other_param:
            cmd += (' ' + self.other_param)

        self.cellranger_count.logger.info(cmd)
        with open(self.cmd_line, 'w') as f:
            f.write(cmd)
        
        cwd = os.getcwd()
        os.chdir(self.outdir)
        subprocess.check_call(cmd, shell=True)
        # change dir back to avoid can not find '03.assemble/stat.txt' error
        os.chdir(cwd)

        cmd = (
            f'cp {self.outdir}/{self.sample}/outs/web_summary.html {self.outdir}/..'
        )
        subprocess.check_call(cmd, shell=True)

    def run(self):
        self.cellranger_count()


def cellranger(args):
    with Cellranger(args) as runner:
        runner.run()


def get_opts_cellranger(parser, sub_program):
    parser.add_argument('--ref_path', help='reference path for cellranger')
    parser.add_argument('--soft_path', help='soft path for cellranger')
    parser.add_argument('--other_param', help='Other cellranger parameters.', default="")
    parser.add_argument('--mem', help='memory(G)', default=10)
    if sub_program:
        s_common(parser)
        parser.add_argument('--fqs_dir', help='fastq dir after convert', required=True)
    return parser


import logging
import os
from collections import defaultdict

import pandas as pd
import pysam

from celescope.tools.utils import add_log
from celescope.tools.step import s_common


def genDict(dim=3):
    if dim == 1:
        return defaultdict(int)
    else:
        return defaultdict(lambda: genDict(dim - 1))


def sum_virus(validated_barcodes, virus_bam,
              out_read_count_file, out_umi_count_file):
    # process bam
    samfile = pysam.AlignmentFile(virus_bam, "rb")
    count_dic = genDict(dim=3)
    for read in samfile:
        tag = read.reference_name
        attr = read.query_name.split('_')
        barcode = attr[0]
        umi = attr[1]
        if barcode in validated_barcodes:
            count_dic[barcode][tag][umi] += 1

    # write dic to pandas df
    rows = []
    for barcode in count_dic:
        for tag in count_dic[barcode]:
            for umi in count_dic[barcode][tag]:
                rows.append([barcode, tag, umi, count_dic[barcode][tag][umi]])
    if len(rows) == 0:
        logging.warning("No cell virus UMI found!")

    df_read = df_read = pd.DataFrame(
        rows,
        columns=[
            "barcode",
            "tag",
            "UMI",
            "read_count"])
    df_read.to_csv(out_read_count_file, sep="\t", index=False)

    df_umi = df_read.groupby(["barcode", "tag"]).agg({"UMI": "count"})
    df_umi.to_csv(out_umi_count_file, sep="\t")


@add_log
def count_virus(args):

    # 检查和创建输出目录
    if not os.path.exists(args.outdir):
        os.system('mkdir -p %s' % (args.outdir))

    # read barcodes
    df_barcodes = pd.read_csv(args.barcode_file, header=None)
    validated_barcodes = list(df_barcodes.iloc[:, 0])

    # count virus
    out_read_count_file = args.outdir + "/" + args.sample + "_virus_read_count.tsv"
    out_umi_count_file = args.outdir + "/" + args.sample + "_virus_UMI_count.tsv"
    sum_virus(
        validated_barcodes,
        args.virus_bam,
        out_read_count_file,
        out_umi_count_file)


def get_opts_count_virus(parser, sub_program):
    if sub_program:
        s_common(parser)
        parser.add_argument('--virus_bam', required=True)
        parser.add_argument('--barcode_file', required=True)



from celescope.rna.__init__ import __ASSAY__
from celescope.tools.multi import Multi


class Multi_rna(Multi):
    """
    ## Usage
    ```
        multi_rna\\
        --mapfile ./rna.mapfile\\
        --genomeDir /SGRNJ/Public/Database/genome/homo_mus\\
        --thread 8\\
        --mod shell
    ```
    Work for both single cell RNA-Seq and single nuclei RNA-Seq.
    """


def main():
    multi = Multi_rna(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()



import os
import subprocess
import pysam
import operator
import random
from xopen import xopen
from collections import defaultdict

from celescope.tools import utils
from celescope.tools.step import Step, s_common


# Template Swithcing Oligos Sequence. [16-bp cell barcode][10/12-bp UMI][TSO].
TSO = "TTTCTTATATGGG"

# Path of Whitelist File in Cellranger Directory.
WHITELIST_10X_PATH = [
    "/lib/python/cellranger/barcodes",
    "/cellranger-cs/3.0.2/lib/python/cellranger/barcodes",
]

# Chemistry: [V2/V3 whitelist, 10X umi length].
CHEMISTRY_DICT = {
    'V2': ['737K-august-2016.txt', 10],
    'V3': ['3M-february-2018.txt.gz', 12],
}


class Convert(Step):
    """
    ##Features

    - Convert barcodes and UMI to 10X format.

    Output

    - `02.convert/barcode_correspond.txt` Recording barcodes correspondence.

    - `02.convert/{sample}_S1_L001_R1_001.fastq.gz` New R1 reads as cellranger input.

    - `02.convert/{sample}_S1_L001_R2_001.fastq.gz` New R2 reads as cellranger input.
    
    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)
        self.fq2 = args.fq2
        self.whitelist_suffix = CHEMISTRY_DICT[args.tenX_chemistry][0]
        self.UMI_10X_LEN = CHEMISTRY_DICT[args.tenX_chemistry][-1]
        
        if not os.path.islink(args.soft_path):
            raise Exception("Error soft_path\nCorrect Format:\n/home/soft/cellranger-7.1.0/cellranger")

        self.version = os.path.dirname(args.soft_path).split('/')[-1].split('-')[-1]
        if int(self.version.split('.')[0]) < 4:
            self.whitelist_10X_file = os.path.dirname(args.soft_path) + f'{WHITELIST_10X_PATH[1]}/{self.whitelist_suffix}'
        else:
            self.whitelist_10X_file = os.path.dirname(args.soft_path) + f'{WHITELIST_10X_PATH[0]}/{self.whitelist_suffix}'

        self.whitelist_10X_fh = xopen(self.whitelist_10X_file, 'r')
        self.sgr_tenX = {}

        # out
        self.out_fq1_file = f'{self.outdir}/{self.sample}_S1_L001_R1_001.fastq.gz'
        self.out_fq2_file = f'{self.outdir}/{self.sample}_S1_L001_R2_001.fastq.gz'
        self.barcode_convert_json = f'{self.outdir}/barcode_convert.json'

    @utils.add_log
    def gen_sgr_tenX_dict(self):
        count_dict = defaultdict(int)
        
        with pysam.FastxFile(self.fq2) as fq2_fh:
            for entry in fq2_fh:
                sgr_barcode = entry.name.split('_')[0]
                count_dict[sgr_barcode] += 1

        count_dict = dict(sorted(count_dict.items(), key=operator.itemgetter(1), reverse=True))
        
        for sgr_barcode in count_dict:
            self.sgr_tenX[sgr_barcode] = self.whitelist_10X_fh.readline().strip()

        # Add invalid barcode
        for sgr_barcode, barcode_10X in self.sgr_tenX.items():
            if barcode_10X == '':
                self.sgr_tenX[sgr_barcode] = "AAAA" + ''.join(random.choice("ATCG") for _ in range(12))
        
    @utils.add_log
    def write_fq1(self):      
        out_fq1 = xopen(self.out_fq1_file, 'w')

        with pysam.FastxFile(self.fq2) as fq2_fh:
            for entry in fq2_fh:
                name = entry.name
                attrs = name.split('_')
                sgr_barcode, sgr_umi = attrs[0], attrs[1]
                new_seq1, new_qual1 = self.convert_seq(sgr_barcode, sgr_umi)
                out_fq1.write(f'@{name}\n{new_seq1}\n+\n{new_qual1}\n')

        out_fq1.close()

    @utils.add_log
    def gzip_fq2(self):
        cmd = f'gzip -c {self.fq2} > {self.out_fq2_file}'
        subprocess.check_call(cmd, shell=True)
   
    def convert_seq(self, barcode_sgr, umi_sgr):
        """
        Convert sgr barcode to 10X barcode; change length of sgr UMI to UMI_10X_LEN

        Args:
            barcode_sgr: str
            umi_sgr: str
        Returns:
            new_seq1: str
            new_qual1: str
        """
        barcode_10X = self.sgr_tenX[barcode_sgr]

        umi_len_sgr = len(umi_sgr)
        if umi_len_sgr > self.UMI_10X_LEN:
            umi_10X = umi_sgr[:self.UMI_10X_LEN]
        elif umi_len_sgr < self.UMI_10X_LEN:
            umi_10X = umi_sgr + 'C' * (self.UMI_10X_LEN - umi_len_sgr)
        else:
            umi_10X = umi_sgr

        new_seq1 = barcode_10X + umi_10X + TSO
        new_qual1 = 'F' * len(new_seq1)

        return new_seq1, new_qual1

    @utils.add_log
    def dump_tenX_sgr_barcode_json(self):
        tenX_sgr = {}
        for sgr, tenX in self.sgr_tenX.items():
            tenX_sgr[tenX] = sgr

        utils.dump_dict_to_json(tenX_sgr, self.barcode_convert_json)

    def run(self):
        self.gen_sgr_tenX_dict()
        self.write_fq1()
        self.gzip_fq2()
        self.dump_tenX_sgr_barcode_json()

def convert(args):
    step_name = 'convert'
    with Convert(args, step_name) as runner:
        runner.run()


def get_opts_convert(parser, sub_program):
    parser.add_argument('--soft_path', help='soft path for cellranger', required=True)
    parser.add_argument(
        '--tenX_chemistry',
        help='10X chemistry version, V2 or V3 for scRNA, V2 for VDJ',
        choices=['V2', 'V3'],
        default='V2')
    if sub_program:
        s_common(parser)
        parser.add_argument('--fq2', help='R2 read file', required=True)
    return parser



import re
import collections
import csv
import sys

from celescope.tools import utils
from celescope.tools.matrix import Features

PATTERN = re.compile(r'(\S+?)\s*"(.*?)"')

class GeneIdNotFound(Exception):
    pass

class GtfParser:
    def __init__(self, gtf_fn):
        self.gtf_fn = gtf_fn
        self.gene_id = []
        self.gene_name = []
        self.id_name = {}

    def get_properties_dict(self, properties_str):
        """
        allow no space after semicolon
        """
        
        if isinstance(properties_str, dict):
            return properties_str

        properties = collections.OrderedDict()
        attrs = properties_str.split(';')
        for attr in attrs:
            if attr:
                m = re.search(PATTERN, attr)
                if m:
                    key = m.group(1)
                    key = key.strip()
                    value = m.group(2)
                    value = value.strip()
                    properties[key] = value

        return properties

    def gtf_reader_iter(self):
        """
        Yield:
            row: str
            is_comment: bool
            annotation: str; exon, gene, etc
            properties: dict; gtf properties in the last column(9th col)
        """
        with utils.generic_open(self.gtf_fn, mode='rt') as f:
            reader = csv.reader(f, delimiter='\t')
            for i, row in enumerate(reader, start=1):
                if len(row) == 0:
                    continue
                if row[0].startswith('#'):
                    yield row, True, None, None
                    continue

                if len(row) != 9:
                    sys.exit(f"Invalid number of columns in GTF line {i}: {row}\n")

                if row[6] not in ['+', '-']:
                    sys.exit(f"Invalid strand in GTF line {i}: {row}\n")
                
                properties = self.get_properties_dict(row[8])
                annotation = row[2]
                if annotation == 'exon':
                    if 'gene_id' not in properties:
                        raise GeneIdNotFound(f"Property 'gene_id' not found in GTF line {i}: {row}\n")

                yield row, False, annotation, properties

    @utils.add_log
    def get_id_name(self):
        """
        get gene_id:gene_name from gtf file
            - one gene_name with multiple gene_id: allowed.
            - one gene_id with multiple gene_name: error.
            - duplicated (gene_name, gene_id): ignore duplicated records and print a warning.
            - no gene_name: gene_id will be used as gene_name.

        """
        for _row, _is_comment, annotation, properties in self.gtf_reader_iter():
            if annotation == 'gene':
                gene_id = properties['gene_id']
                if 'gene_name' not in properties:
                    gene_name = gene_id
                else:
                    gene_name = properties['gene_name']

                if gene_id in self.id_name:
                    assert self.id_name[gene_id] == gene_name, (
                        'one gene_id with multiple gene_name '
                        f'gene_id: {gene_id}, '
                        f'gene_name this line: {gene_name}'
                        f'gene_name previous line: {self.id_name[gene_id]}'
                    )
                    self.get_id_name.logger.warning(
                        'duplicated (gene_id, gene_name)'
                        f'gene_id: {gene_id}, '
                        f'gene_name {gene_name}'
                    )
                else:
                    self.gene_id.append(gene_id)
                    self.gene_name.append(gene_name)
                    self.id_name[gene_id] = gene_name
                    
        return self.id_name
                    
    def get_features(self):
        """
        Returns:
            Features object
        """
        if not self.gene_id:
            sys.exit("Empty self.gene_id. Run self.get_id_name first.")
        features = Features(self.gene_id, self.gene_name)
        return features


class GtfBuilder:
    def __init__(self, in_gtf_fn, out_gtf_fn, attributes):
        self.in_gtf_fn = in_gtf_fn
        self.out_gtf_fn = out_gtf_fn
        self.attributes = attributes

    @utils.add_log
    def build_gtf(self):
        """
        Fix genes without gene annotation
        Filter gene biotypes
        """
        self.build_gtf.logger.info("Writing filtered GTF file...")
        gp = GtfParser(self.in_gtf_fn)

        exons = collections.defaultdict(dict)
        genes = set()

        with open(self.out_gtf_fn, 'w') as f:
            writer = csv.writer(f, delimiter='\t', quoting=csv.QUOTE_NONE, quotechar='')
            for row, is_comment, annotation, properties in gp.gtf_reader_iter():
                if is_comment:
                    writer.writerow(row)
                    continue
                
                if annotation == 'gene':
                    genes.add(properties['gene_id'])
                elif annotation == 'exon':
                    gene_id = properties['gene_id']
                    if gene_id not in genes:
                        seqID = row[0]
                        start, end = int(row[3]), int(row[4])
                        strand = row[6]
                        if gene_id not in exons:
                            exons[gene_id]['strand'] = strand
                            exons[gene_id]['start'] = start
                            exons[gene_id]['end'] = end
                            exons[gene_id]['properties'] = row[8]
                            exons[gene_id]['seqID'] = seqID
                        else:
                            if strand != exons[gene_id]['strand']:
                                self.build_gtf.logger.warning(f'Error: gene {gene_id} is on both strand!\nline: {row}')
                                continue
                            exons[gene_id]['start'] = min(start, exons[gene_id]['start'])
                            exons[gene_id]['end'] = max(end, exons[gene_id]['end'])


                remove = False
                for key, value in properties.items():
                    if key in self.attributes and value not in self.attributes[key]:
                        remove = True

                if not remove:
                    writer.writerow(row)


            for gene_id, vals in exons.items():
                if gene_id not in genes:
                    seqID, strand, start, end, properties = vals['seqID'], vals['strand'], str(vals['start']), str(vals['end']),vals['properties']
                    row = [seqID, 'added', 'gene', start, end, '.', strand, '.', properties]
                    writer.writerow(row)





        




from celescope.tools.multi import Multi
from celescope.convert10X.__init__ import __ASSAY__


class Multi_convert10X(Multi):
    """

    ## Usage
    Running cellranger-count for sgr sc-RNA data:
    ```
    conda activate celescope
    multi_convert10X \\
        --mapfile  test.mapfile \\
        --thread 8 \\
        --ref_path "/soft/cellranger/refdata-gex-GRCh38-2020-A" \\
        --soft_path "/soft/cellranger/cellranger-6.1.2/cellranger" \\
        --mod shell 
    ``` 
    Converting sgr data to 10X format:
    ```
    conda activate celescope
    multi_convert10X \\
        --mapfile  test.mapfile \\
        --thread 8 \\
        --soft_path "/soft/cellranger/cellranger-6.1.2/cellranger" \\
        --steps_run sample,barcode,convert \\
        --mod shell 
    ``` 
    """

    def convert(self, sample):
        step = 'convert'
        cmd_line = self.get_cmd_line(step, sample)
        fq2 = f'{self.outdir_dic[sample]["barcode"]}/{sample}_2.fq'
        cmd = (
            f'{cmd_line} '
            f'--fq2 {fq2} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)

    def cellranger(self, sample):
        step = 'cellranger'
        cmd_line = self.get_cmd_line(step, sample)
        fqs_dir = f'{self.outdir_dic[sample]["convert"]}'
        cmd = (
            f'{cmd_line} '
            f'--fqs_dir {fqs_dir} '
        )
        self.process_cmd(cmd, step, sample, m=self.args.mem, x=self.args.thread)

        
def main():
    multi = Multi_convert10X(__ASSAY__)
    multi.run()
    

if __name__ == '__main__':
    main()




from celescope.tools.mkref import Mkref, super_opts

class Mkref_fusion(Mkref):
    """
    ## Features
    - Create a fusion genome directory.

    ## Output

    - STAR genome index files
    - Genome config file

    ## Usage
    ```
    celescope fusion mkref \\
    --genome_name {genome_name} \\
    --fasta fusion.fasta \\
    --fusion_pos fusion_pos.txt \\
    ```
    """

    def run(self):
        super().run()
        self.build_star_index()

    @staticmethod
    def parse_genomeDir(genomeDir):
        return Mkref.parse_genomeDir(genomeDir, files=('fusion_pos',))



def mkref(args):
    genome_type = 'fusion'
    with Mkref_fusion(genome_type, args, files=('fusion_pos',), ) as runner:
        runner.run()


def get_opts_mkref(parser, sub_program):
    super_opts(parser, sub_program)
    if sub_program:
        parser.add_argument(
            "--fusion_pos",
            help="""
fusion position file. A two column tab-delimited text file with header.
"pos" is the end postion of the first gene(1-based).
e.g.  
name\tpos  
PML_3\t183  
PML_4\t254  
PML_5\t326  
PML_6\t204   
""",
            required=True,)




import re
import subprocess
import os

from celescope.tools import utils
from celescope.tools.mkref import Mkref
from celescope.tools.step import s_common
from celescope.__init__ import HELP_DICT
from celescope.tools.step import Step


@utils.add_log
def get_star_cmd(args, input_file, output_prefix):
    """
    output sam format to improve speed
    """
    cmd = (
        f'STAR '
        f'--runThreadN {args.thread} '
        f'--genomeDir {args.genomeDir} '
        f'--outFilterMultimapNmax {args.outFilterMultimapNmax} '
        f'--outSAMtype BAM Unsorted '
        f'--outFilterMatchNmin {args.outFilterMatchNmin} '
        f'{args.STAR_param} '
        f'--readFilesIn {input_file} '
        f'--outFileNamePrefix {output_prefix}_ '
    )
    get_star_cmd.logger.info(cmd)
    return cmd


class Star_mixin(Step):
    """
    Mixin class for STAR
    """

    def __init__(self, args, add_prefix=None, display_title=None):
        super().__init__(args, display_title)

        # parse
        self.genome = Mkref.parse_genomeDir(args.genomeDir)
        self.stat_prefix = 'Reads'
        if getattr(args, "consensus_fq", False):
            self.stat_prefix = 'UMIs'

        # out
        if add_prefix:
            self.out_prefix += f'_{add_prefix}'
        self.STAR_map_log = f'{self.out_prefix}_Log.final.out'
        self.unsort_STAR_bam = f'{self.out_prefix}_Aligned.out.bam'
        self.STAR_bam = f'{self.out_prefix}_Aligned.sortedByCoord.out.bam'

    @utils.add_log
    def STAR(self):
        input_file = self.args.fq
        output_prefix = self.out_prefix
        cmd = get_star_cmd(self.args, input_file, output_prefix)

        if self.args.out_unmapped:
            cmd += '--outReadsUnmapped Fastx'
        if self.args.fq[-3:] == ".gz":
            cmd += '--readFilesCommand zcat'
        if self.args.STAR_param:
            cmd += (" " + self.args.STAR_param)
        self.STAR.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    def run(self):
        self.STAR()
        self.add_star_metrics()
        self.sort_bam()
        self.index_bam()
        self.remove_unsort_bam()

    @utils.add_log
    def sort_bam(self):
        utils.sort_bam(
            self.unsort_STAR_bam,
            self.STAR_bam,
            threads=self.thread,
        )

    @utils.add_log
    def index_bam(self):
        utils.index_bam(self.STAR_bam)

    @utils.add_log
    def remove_unsort_bam(self):
        os.remove(self.unsort_STAR_bam)

    def add_star_metrics(self):
        """
        step metrics
        """

        with open(self.STAR_map_log, 'r') as map_log:
            # number amd percent
            unique_reads_list = []
            multi_reads_list = []
            total_reads = 0
            for line in map_log:
                if line.strip() == '':
                    continue
                if re.search(r'Uniquely mapped reads', line):
                    unique_reads_list.append(line.strip().split()[-1])
                if re.search(r'of reads mapped to too many loci', line):
                    multi_reads_list.append(line.strip().split()[-1])
                if re.search(r'Number of input reads', line):
                    total_reads = int(line.strip().split()[-1])

        unique_reads = int(unique_reads_list[0])
        multi_reads = int(multi_reads_list[0])

        self.add_metric(
            name='Genome',
            value=self.genome['genome_name'],
        )
        self.add_metric(
            name=f'Uniquely Mapped {self.stat_prefix}',
            value=unique_reads,
            total=total_reads,
            help_info='reads that mapped uniquely to the genome'
        )
        self.add_metric(
            name=f'Multi-Mapped {self.stat_prefix}',
            value=multi_reads,
            total=total_reads,
            help_info='reads that mapped to multiple locations in the genome'
        )


def get_opts_star_mixin(parser, sub_program):
    parser.add_argument(
        '--genomeDir',
        help=HELP_DICT['genomeDir'],
    )
    parser.add_argument(
        '--outFilterMatchNmin',
        help="""Alignment will be output only if the number of matched bases 
is higher than or equal to this value.""",
        default=50,
    )
    parser.add_argument(
        '--out_unmapped',
        help='Output unmapped reads.',
        action='store_true'
    )
    parser.add_argument('--STAR_param', help=HELP_DICT['additional_param'], default="")
    parser.add_argument(
        '--outFilterMultimapNmax',
        help='Default `1`. How many places are allowed to match a read at most.',
        default=1
    )
    parser.add_argument(
        '--starMem',
        help='Default `30`. Maximum memory that STAR can use.',
        default=30
    )
    if sub_program:
        parser.add_argument('--fq', help="Required. R2 fastq file.", required=True)
        parser.add_argument("--consensus_fq", action='store_true',
                            help="A indicator that the input fastq has been consensused.")
        parser = s_common(parser)




import os
import pathlib
from collections import defaultdict
from itertools import groupby
import subprocess
import unittest

import pandas as pd
import pysam

from celescope.rna.mkref import Mkref_rna
from celescope.tools.step import Step, s_common
from celescope.tools import utils
from celescope.__init__ import HELP_DICT
from celescope.tools import reference


GTF_TYPES = ['exon','gene']

def add_tag(seg, id_name, gene_correct_umi):
    """
    Args:
        seg: pysam bam segment
        id_name: {gene_id: gene_name}
        correct_dict: {low_seq: high_seq}

    Returns:
        seg with tag added

    Tags:
        CB: cell barcode
        UB: error-corrected UMI
        UR: original UMI
        GN: gene name
        GX: gene_id

    """
    attr = seg.query_name.split('_')
    barcode = attr[0]
    ur = ub = attr[1]

    # assign to some gene
    if seg.has_tag('XT'):
        gene_id = seg.get_tag('XT')
        # if multi-mapping reads are included in original bam,
        # there are multiple gene_ids
        if ',' in gene_id:
            gene_name = [id_name[i] for i in gene_id.split(',')]
            gene_name = ','.join(gene_name)
        else:
            gene_name = id_name[gene_id]
        seg.set_tag(tag='GN', value=gene_name, value_type='Z')
        seg.set_tag(tag='GX', value=gene_id, value_type='Z')

        if gene_id in gene_correct_umi and ur in gene_correct_umi[gene_id]:
            ub = gene_correct_umi[gene_id][ur]
    
    seg.set_tag(tag='CB', value=barcode, value_type='Z')
    seg.set_tag(tag='UB', value=ub, value_type='Z')
    seg.set_tag(tag='UR', value=ur, value_type='Z')

    return seg


def correct_umi(umi_dict, percent=0.1):
    """
    Correct umi_dict in place.
    Args:
        umi_dict: {umi_seq: umi_count}
        percent: if hamming_distance(low_seq, high_seq) == 1 and
            low_count / high_count < percent, merge low to high.
        return_dict: if set, return correct_dict = {low_seq:high_seq}
    Returns:
        n_corrected_umi: int
        n_corrected_read: int
    """
    n_corrected_umi = 0
    n_corrected_read = 0
    dic = {}

    # sort by value(UMI count) first, then key(UMI sequence)
    umi_arr = sorted(
        umi_dict.items(), key=lambda kv: (kv[1], kv[0]), reverse=True)
    while True:
        # break when only highest in umi_arr
        if len(umi_arr) == 1:
            break
        umi_low = umi_arr.pop()
        low_seq = umi_low[0]
        low_count = umi_low[1]

        for umi_kv in umi_arr:
            high_seq = umi_kv[0]
            high_count = umi_kv[1]
            if float(low_count / high_count) > percent:
                break
            if utils.hamming_distance(low_seq, high_seq) == 1:
                n_low = umi_dict[low_seq]
                n_corrected_umi += 1
                n_corrected_read += n_low
                # merge
                umi_dict[high_seq] += n_low
                dic[low_seq] = high_seq
                del (umi_dict[low_seq])
                break

    return n_corrected_umi, n_corrected_read, dic


def discard_read(gene_umi_dict):
    """
    If two or more groups of reads have the same barcode and UMI, but different gene annotations, the gene annotation with the most supporting reads is kept for UMI counting, and the other read groups are discarded. In case of a tie for maximal read support, all read groups are discarded, as the gene cannot be confidently assigned.

    Returns:
        discarded_umi: set. umi with tie read count
        umi_gene_dict: {umi_seq: {gene_id: read_count}}
    """

    discard_umi = set()
    umi_gene_dict = defaultdict(lambda: defaultdict(int))
    for gene_id in gene_umi_dict:
        for umi in gene_umi_dict[gene_id]:
            umi_gene_dict[umi][gene_id] += gene_umi_dict[gene_id][umi]
    
    for umi in umi_gene_dict:
        max_read_count = max(umi_gene_dict[umi].values())
        gene_id_max = [gene_id for gene_id, read_count in umi_gene_dict[umi].items() if read_count==max_read_count]

        if len(gene_id_max) > 1:
            discard_umi.add(umi)
        else:
            gene_id = gene_id_max[0]
            umi_gene_dict[umi] = {gene_id: umi_gene_dict[umi][gene_id]}

    return discard_umi, umi_gene_dict


class FeatureCounts(Step):

    """
    ## Features
    - Assigning uniquely mapped reads to genomic features with FeatureCounts.
    ## Output
    - `{sample}` Numbers of reads assigned to features (or meta-features).
    - `{sample}_summary` Stat info for the overall summrization results, including number of 
    successfully assigned reads and number of reads that failed to be assigned due to 
    various reasons (these reasons are included in the stat info).
    - `{sample}_aligned_sortedByCoord_addTag.bam` featureCounts output BAM, 
    sorted by coordinates;BAM file contains tags as following(Software Version>=1.1.8):
        - CB cell barcode
        - UB UMI
        - GN gene name
        - GX gene id
    - `{sample}_name_sorted.bam` featureCounts output BAM, sorted by read name.
    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        # set
        self.gtf = Mkref_rna.parse_genomeDir(self.args.genomeDir)['gtf']
        gp = reference.GtfParser(self.gtf)
        self.id_name = gp.get_id_name()

        # stats
        self.feature_log_dict = defaultdict(dict)
        self.n_corrected_read = 0
        self.n_corrected_umi = 0

        # out
        self.count_detail_file = f'{self.out_prefix}_count_detail.txt'
        input_basename = os.path.basename(self.args.input)
        self.featureCounts_bam = f'{self.outdir}/{input_basename}.featureCounts.bam'
        self.name_sorted_bam = f'{self.out_prefix}_name_sorted.bam'
        self.add_tag_sam = f'{self.out_prefix}_addTag.sam'
        self.out_bam = f'{self.out_prefix}_aligned_sortedByCoord_addTag.bam'


    @staticmethod
    def read_log(log_file):
        """
        Args:
            log_file: featureCounts log summary file
        Returns:
            log_dict: {'Assigned': 123, ...}
        """
        # skip first line
        df = pd.read_csv(log_file, sep='\t', header=None, names=['name', 'value'], skiprows=1)
        log_dict = df.set_index('name')['value'].to_dict()
        return log_dict


    @utils.add_log
    def run_featureCounts(self, outdir, gtf_type):
        cmd = (
            'featureCounts '
            '-s 1 '
            f'-a {self.gtf} '
            f'-o {outdir}/{self.sample} '  
            '-R BAM '
            f'-T {self.args.thread} '
            f'-t {gtf_type} '
            f'{self.args.input} '
            '2>&1 '
        )
        if self.args.featureCounts_param:
            cmd += (" " + self.args.featureCounts_param)
        self.run_featureCounts.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    def run_get_log(self):
        tmp_dir = f'{self.outdir}/tmp/'
        for gtf_type in GTF_TYPES:
            outdir = f'{tmp_dir}/{gtf_type}'
            pathlib.Path(outdir).mkdir(parents=True, exist_ok=True)
            log_file = f'{outdir}/{self.sample}.summary'
            
            self.run_featureCounts(outdir, gtf_type)
            self.feature_log_dict[gtf_type] = FeatureCounts.read_log(log_file)

            if gtf_type == self.args.gtf_type:
                cmd = f'mv {outdir}/* {self.outdir}/ '
                subprocess.check_call(cmd, shell=True)

        cmd = f'rm -r {tmp_dir} '
        subprocess.check_call(cmd, shell=True)


    def remove_temp_file(self):
        os.remove(self.name_sorted_bam)
        os.remove(self.add_tag_sam)
        os.remove(self.featureCounts_bam)

    def run(self):
        self.run_get_log()
        self.add_metrics()
        utils.sort_bam(
            input_bam=self.featureCounts_bam,
            output_bam=self.name_sorted_bam, 
            by='name')
        self.get_count_detail_add_tag()
        utils.sort_bam(
            input_bam=self.add_tag_sam,
            output_bam=self.out_bam)
        self.remove_temp_file()

    @utils.add_log
    def add_metrics(self):
        total = sum(self.feature_log_dict['exon'].values())

        Assigned_exon = self.feature_log_dict['exon']['Assigned']
        Assigned_intergenic = self.feature_log_dict['gene']['Unassigned_NoFeatures']
        """
        https://academic.oup.com/nargab/article/2/3/lqaa073/5910008
        Approximately 15% of genes had exon counts that were greater than genebody counts (by a median value of eight counts). This was due to our conservative approach of excluding reads that overlapped features in multiple genes during the read summarization step by featureCounts using the argument allowMultiOverlap=FALSE. Under this strategy, some reads were counted towards the exon count set but not the genebody count set. This happens when a read overlaps the exon in one gene and the intron of another gene—it is counted towards exon counts but not genebody counts due to its overlap of multiple genebodies but not multiple exons.
        """
        Unassigned_ambiguity = self.feature_log_dict['exon']['Unassigned_Ambiguity']
        Assigned_intron = total - Assigned_exon - Assigned_intergenic - Unassigned_ambiguity           
       
        
        self.add_metric(
            name='Feature Type',
            value=self.args.gtf_type.capitalize(),
            help_info='Specified by `--gtf_type`. For snRNA-seq, you need to add `--gtf_type gene` to include reads mapped to intronic regions. Staring from CeleScope v1.12.0, the default value of gtf_type is changed from `exon` to `gene`.'
        )
        self.add_metric(
            name='Reads Assigned To Exonic Regions',
            value=Assigned_exon,
            total=total,
            help_info='Reads that can be successfully assigned to exonic regions'
        )
        self.add_metric(
            name='Reads Assigned To Intronic Regions',
            value=Assigned_intron,
            total=total,
            help_info='Reads that can be successfully assigned to intronic regions'
        )
        self.add_metric(
            name='Reads Assigned To Intergenic Regions',
            value=Assigned_intergenic,
            total=total,
            help_info='Reads that can be successfully assigned to intergenic regions'
        )
        self.add_metric(
            name='Reads Unassigned Ambiguity',
            value=Unassigned_ambiguity,
            total=total,
            help_info='Alignments that overlap two or more features'
        )

    @utils.add_log
    def get_count_detail_add_tag(self):
        """
        bam to detail table
        must be used on name_sorted bam
        Output file:
            - count_detail_file
            - bam with tag(remain name sorted)
        """
        save = pysam.set_verbosity(0)
        inputFile = pysam.AlignmentFile(self.name_sorted_bam, "rb")
        outputFile = pysam.AlignmentFile(self.add_tag_sam, 'w', header=inputFile.header)
        pysam.set_verbosity(save)

        with open(self.count_detail_file, 'wt') as fh1:
            fh1.write('\t'.join(['Barcode', 'geneID', 'UMI', 'read', 'duplicate']) + '\n')

            def keyfunc(x):
                return x.query_name.split('_', 1)[0]
            for _, g in groupby(inputFile, keyfunc):
                gene_umi_dict = defaultdict(lambda: defaultdict(int))
                gene_umi_pos_cigar = utils.genDict(dim=4, valType=int)
                segs = []
                for seg in g:
                    segs.append(seg)
                    (barcode, umi) = seg.query_name.split('_')[:2]
                    if not seg.has_tag('XT'):
                        continue
                    gene_id = seg.get_tag('XT')
                    gene_umi_dict[gene_id][umi] += 1
                    gene_umi_pos_cigar[gene_id][umi][seg.reference_start][seg.cigarstring] += 1

                gene_correct_umi = dict()
                for gene_id in gene_umi_dict:
                    n_corrected_umi, n_corrected_read, correct_dict = correct_umi(gene_umi_dict[gene_id])
                    gene_correct_umi[gene_id] = correct_dict
                    self.n_corrected_read += n_corrected_read
                    self.n_corrected_umi += n_corrected_umi

                    # also correct umi in gene_umi_pos_cigar
                    for low_seq, high_seq in correct_dict.items():
                        for ref_start in gene_umi_pos_cigar[gene_id][low_seq]:
                            for cigar in gene_umi_pos_cigar[gene_id][low_seq][ref_start]:
                                gene_umi_pos_cigar[gene_id][high_seq][ref_start][cigar] += 1
                        del gene_umi_pos_cigar[gene_id][low_seq]                    

                # output
                for gene_id in gene_umi_dict:
                    for umi in gene_umi_dict[gene_id]:
                        dup_list = []
                        for pos in gene_umi_pos_cigar[gene_id][umi]:
                            for cigar in gene_umi_pos_cigar[gene_id][umi][pos]:
                                dup_list.append(str(gene_umi_pos_cigar[gene_id][umi][pos][cigar]))
                        duplicate = ','.join(dup_list)
                        fh1.write(f'{barcode}\t{gene_id}\t{umi}\t{gene_umi_dict[gene_id][umi]}\t{duplicate}\n')

                for seg in segs:
                    outputFile.write(add_tag(seg, self.id_name, gene_correct_umi))

        self.add_metric('n_corrected_read', self.n_corrected_read, show=False)
        self.add_metric('n_corrected_umi', self.n_corrected_umi, show=False)

        inputFile.close()
        outputFile.close()


@utils.add_log
def featureCounts(args):
    with FeatureCounts(args) as runner:
        runner.run()


def get_opts_featureCounts(parser, sub_program):
    parser.add_argument(
        '--gtf_type',
        help='Specify feature type in GTF annotation',
        default='gene',
        choices=['exon', 'gene'],
    )
    parser.add_argument('--genomeDir', help=HELP_DICT['genomeDir'])
    parser.add_argument('--featureCounts_param', help=HELP_DICT['additional_param'], default="")

    if sub_program:
        parser.add_argument('--input', help='Required. BAM file path.', required=True)
        parser = s_common(parser)
    return parser


class featureCounts_test(unittest.TestCase):
    def test_correct_umi(self):
        dic = {
            "apple1": 2,
            "apple2": 30,
            "bears1": 5,
            "bears2": 10,
            "bears3": 100,
            "ccccc1": 20,
            "ccccc2": 199,
        }
        n_corrected_umi, n_corrected_read, _ = correct_umi(dic)
        dic_after_correct = {
            'ccccc1': 20,
            'apple2': 32,
            'bears3': 115,
            'ccccc2': 199,
        }
        self.assertEqual(dic, dic_after_correct)
        self.assertEqual(n_corrected_umi, 3)
        self.assertEqual(n_corrected_read, 2 + 5 + 10)

if __name__ == "__main__":
    unittest.main()


#!/usr/bin/env python
#
# Copyright (c) 2018 10X Genomics, Inc. All rights reserved.
#

"""
Simple Good-Turing estimator.
Based on S implementation in
  William A. Gale & Geoffrey Sampson (1995) Good-turing frequency estimation without tears,
  Journal of Quantitative Linguistics, 2:3, 217-237, DOI: 10.1080/09296179508590051
"""

import numpy as np
import scipy.stats as sp_stats


class SimpleGoodTuringError(Exception):
    pass


def _averaging_transform(r, nr):
    d = np.concatenate((np.ones(1, dtype=int), np.diff(r)))
    dr = np.concatenate((
        0.5 * (d[1:] + d[0:-1]),
        np.array((d[-1],), dtype=float),
    ))
    return nr.astype(float)/dr


def _rstest(r, coef):
    return r * np.power(1 + 1/r, 1 + coef)


def simple_good_turing(xr, xnr):
    """Make a Simple Good-Turing estimate of the frequencies.
    Args:
      xr (np.array(int)): Non-zero item frequencies
      xnr (np.array(int)): Non-zero frequencies of frequencies
    Returns:
      (rstar (np.array(float)), p0 (float)):
        rstar: The adjusted non-zero frequencies
        p0: The total probability of unobserved items
    """

    xr = xr.astype(float)
    xnr = xnr.astype(float)

    xN = np.sum(xr*xnr)

    # Get Linear Good-Turing estimate
    xnrz = _averaging_transform(xr, xnr)
    slope, _intercept, _, _, _ = sp_stats.linregress(np.log(xr), np.log(xnrz))

    if slope > -1:
        raise SimpleGoodTuringError(
            "The log-log slope is > -1 (%d); the SGT estimator is not applicable to these data." % slope)

    xrst = _rstest(xr, slope)
    xrstrel = xrst/xr

    # Get traditional Good-Turing estimate
    xrtry = xr == np.concatenate((xr[1:]-1, np.zeros(1)))
    xrstarel = np.zeros(len(xr))
    xrstarel[xrtry] = (xr[xrtry]+1) / xr[xrtry] * np.concatenate((xnr[1:], np.zeros(1)))[xrtry] / xnr[xrtry]

    # Determine when to switch from GT to LGT estimates
    tursd = np.ones(len(xr))
    for i in range(len(xr)):
        if xrtry[i]:
            tursd[i] = float(i+2) / xnr[i] * np.sqrt(xnr[i+1] * (1 + xnr[i+1]/xnr[i]))

    xrstcmbrel = np.zeros(len(xr))
    useturing = True
    for r in range(len(xr)):
        if not useturing:
            xrstcmbrel[r] = xrstrel[r]
        else:
            if np.abs(xrstrel[r]-xrstarel[r]) * (1+r)/tursd[r] > 1.65:
                xrstcmbrel[r] = xrstarel[r]
            else:
                useturing = False
                xrstcmbrel[r] = xrstrel[r]

    # Renormalize the probabilities for observed objects
    sumpraw = np.sum(xrstcmbrel * xr * xnr / xN)

    xrstcmbrel = xrstcmbrel * (1 - xnr[0] / xN) / sumpraw
    p0 = xnr[0]/xN

    return (xr * xrstcmbrel, p0)


def sgt_proportions(frequencies):
    """Use Simple Good-Turing estimate to adjust for unobserved items
    Args:
      frequencies (np.array(int)): Nonzero frequencies of items
    Returns:
      (pstar (np.array(float)), p0 (float)):
        pstar: The adjusted non-zero proportions
        p0: The total probability of unobserved items
    """
    if len(frequencies) == 0:
        raise ValueError("Input frequency vector is empty")
    if np.count_nonzero(frequencies) != len(frequencies):
        raise ValueError("Frequencies must be greater than zero")

    freqfreqs = np.bincount(frequencies)
    assert freqfreqs[0] == 0
    use_freqs = np.flatnonzero(freqfreqs)

    if len(use_freqs) < 10:
        raise SimpleGoodTuringError("Too few non-zero frequency items (%d). Aborting SGT." % len(use_freqs))

    rstar, p0 = simple_good_turing(use_freqs, freqfreqs[use_freqs])

    # rstar contains the smoothed frequencies.
    # Map each original frequency r to its smoothed rstar.
    rstar_dict = dict(zip(use_freqs, rstar))

    rstar_sum = np.sum(freqfreqs[use_freqs] * rstar)
    rstar_i = np.fromiter((rstar_dict[f] for f in frequencies),
                          dtype=float, count=len(frequencies))
    pstar = (1 - p0) * (rstar_i / rstar_sum)

    assert np.isclose(p0 + np.sum(pstar), 1)
    return (pstar, p0)


def test_prosody():
    data = (
        (1, 120),
        (2, 40),
        (3, 24),
        (4, 13),
        (5, 15),
        (6, 5),
        (7, 11),
        (8, 2),
        (9, 2),
        (10, 1),
        (12, 3),
        (14, 2),
        (15, 1),
        (16, 1),
        (17, 3),
        (19, 1),
        (20, 3),
        (21, 2),
        (23, 3),
        (24, 3),
        (25, 3),
        (26, 2),
        (27, 2),
        (28, 1),
        (31, 2),
        (32, 2),
        (33, 1),
        (34, 2),
        (36, 2),
        (41, 3),
        (43, 1),
        (45, 3),
        (46, 1),
        (47, 1),
        (50, 1),
        (71, 1),
        (84, 1),
        (101, 1),
        (105, 1),
        (121, 1),
        (124, 1),
        (146, 1),
        (162, 1),
        (193, 1),
        (199, 1),
        (224, 1),
        (226, 1),
        (254, 1),
        (257, 1),
        (339, 1),
        (421, 1),
        (456, 1),
        (481, 1),
        (483, 1),
        (1140, 1),
        (1256, 1),
        (1322, 1),
        (1530, 1),
        (2131, 1),
        (2395, 1),
        (6925, 1),
        (7846, 1),
    )

    # Computed using R 3.5.1 w/ the Gale S code
    expect_p0 = 0.003883244
    expect_rstar = np.array((
        0.7628079,
        1.706448,
        2.679796,
        3.663988,
        4.653366,
        5.645628,
        6.63966,
        7.634856,
        8.63086,
        9.627446,
        11.62182,
        13.61725,
        14.61524,
        15.61336,
        16.6116,
        18.60836,
        19.60685,
        20.6054,
        22.60264,
        23.60133,
        24.60005,
        25.5988,
        26.59759,
        27.59639,
        30.59294,
        31.59183,
        32.59073,
        33.58964,
        35.58751,
        40.58235,
        42.58035,
        44.57836,
        45.57738,
        46.57641,
        49.57351,
        70.55399,
        83.54229,
        100.5272,
        104.5237,
        120.5097,
        123.507,
        145.4879,
        161.474,
        192.4472,
        198.4421,
        223.4205,
        225.4188,
        253.3947,
        256.3922,
        338.3218,
        420.2514,
        455.2215,
        480.2,
        482.1983,
        1138.636,
        1254.537,
        1320.48,
        1528.302,
        2128.788,
        2392.562,
        6918.687,
        7838.899,
    ))

    xr = np.array([d[0] for d in data], dtype=int)
    xnr = np.array([d[1] for d in data], dtype=int)

    rstar, p0 = simple_good_turing(xr, xnr)

    assert np.abs(p0 - expect_p0) < 1e-9
    assert np.all(np.abs(rstar - expect_rstar) < 1e-3)
    assert np.all((np.abs(rstar - expect_rstar))/expect_rstar < 1e-4)



STEPS = [
    'sample',
    'barcode',
    'cutadapt',
    'star',
    'prep_map',
    "featureCounts",
    "count",
    'analysis',
    'conversion',
    'substitution',
    'replacement',
    'replace_tsne']

__ASSAY__ = 'dynaseq'

IMPORT_DICT = {
    'star': 'celescope.rna',
    'analysis': 'celescope.rna',
}


REMOVE_FROM_MULTI = {
    'barcode',
    'cutadapt',
    'star',
}



#!/usr/bin/env python
# v1.0

import pysam
import os
import subprocess
import numpy as np
import pandas as pd
from collections import Counter
from celescope.tools.step import Step, s_common
from celescope.tools import utils


class Conversion(Step):
    """
    ## Features
    - Get conversion pos in each read.
        - Get snp info. 

    ## Output
    - `{sample}.PosTag.bam` Bam file with conversion info.
    - `{sample}.PosTag.csv` SNP info in csv format.
    """

    def __init__(self, args):
        Step.__init__(self, args)
        # input files
        #self.ifile = os.path.join(args.outdir, args.sample+'.bam')
        self.sample = args.sample
        self.strandednessfile = args.strand
        self.inbam = args.bam
        self.bcfile = args.cell
        self.outdir = args.outdir
        self.thread = args.thread
        self.qual = int(args.basequalilty)

        # output files
        self.outfile_bam = os.path.join(args.outdir, args.sample+'.PosTag.bam')
        self.outfile_csv = os.path.join(args.outdir, args.sample+'.PosTag.csv')

    @utils.add_log
    def run(self):

        # Adding tags
        ContigLocs, AnnoteLocs = self.addTags(self.inbam, self.outfile_bam, self.bcfile, self.strandednessfile, self.qual, self.thread)
        cmd = ['samtools index', self.outfile_bam]
        self.run_cmd(cmd)

        # Obtaining conversion positions
        bam = pysam.AlignmentFile(self.outfile_bam, 'rb')
        #ContigLocs, AnnoteLocs = self.CountConvperPos(bam)

        # Obtaining coverage over conversion position
        ConvsPerPos, CoverofPosWithConvs = self.CountReadConverPerConvPos(bam, ContigLocs)
        A = self.ExportasVcf(ConvsPerPos, CoverofPosWithConvs, AnnoteLocs)
        A['sample'] = self.sample
        # Saving result
        A.to_csv(self.outfile_csv)
        bam.close()

        #cmd = ['rm', self.ifile]
        #self.run_cmd(cmd)
        #cmd = ['rm', self.ifile+'.bai']
        #self.run_cmd(cmd)

    def run_cmd(self, cmd):
        subprocess.call(' '.join(cmd), shell=True)


    #@utils.add_log
    def CountConvperPos_read(self, read, ContigLocs, AnnoteLocs):
        #ContigLocs = {}
        #AnnoteLocs = {}
        try:
            if read.get_tag('ST') == '+':
                locs = read.get_tag('TL')
            else:
                locs = read.get_tag('AL')
            if locs[0] != 0:
                if read.reference_name in ContigLocs:
                    ContigLocs[read.reference_name].extend(locs)
                else:
                    ContigLocs[read.reference_name] = list(locs)
                if read.reference_name not in AnnoteLocs:
                    for i, each in enumerate(locs):
                        if i == 0:
                            AnnoteLocs[read.reference_name] = {each: read.get_tag('GX')}
                        else:
                            AnnoteLocs[read.reference_name][each] = read.get_tag('GX')
                else:
                    for i, each in enumerate(locs):
                        if each not in AnnoteLocs[read.reference_name]:
                            AnnoteLocs[read.reference_name][each] = read.get_tag('GX')
        except (ValueError, KeyError):
            print("Reads Error {}".format(str(read.qurey_name)))
        return ContigLocs, AnnoteLocs

    @utils.add_log
    def CountReadConverPerConvPos(self, bam, ContigLocs):
        ConvsPerPos = {}
        CoverofPosWithConvs = {}
        for key in ContigLocs.keys():
            ConvsPerPos[key] = {}
            ContigLocs[key] = sorted(ContigLocs[key])
            ConvsPerPos[key] = Counter(ContigLocs[key])

            CoverofPosWithConvs[key] = {}
            for key2 in ConvsPerPos[key].keys():
                try:
                    CoverofPosWithConvs[key][key2] = bam.count(key, key2, key2+1)
                except ValueError:
                    continue
        return ConvsPerPos, CoverofPosWithConvs

    @utils.add_log
    def ExportasVcf(self, ConvsPerPos, CoverofPosWithConvs, AnnoteLocs):
        #Chrom, Pos , ConvsPerPs, CoverofPosWithConvs
        order_columns=['chrom', 'pos', 'convs', 'covers', 'posratio','gene_id']
        Outputdf = pd.DataFrame(columns=order_columns)
        for key in ConvsPerPos.keys():
            if len(ConvsPerPos[key])==0: continue
            df = pd.DataFrame.from_dict(ConvsPerPos[key], orient='index')
            df1 = pd.DataFrame.from_dict(CoverofPosWithConvs[key], orient='index')
            df3 = pd.DataFrame.from_dict(AnnoteLocs[key], orient='index')
            #df.index.name = 'pos'
            #df1.index.name = 'pos'
            df.columns = ['convs']
            df1.columns = ['covers']
            df3.columns = ['gene_id']
            df2 = pd.concat([df,df1,df3],axis=1)
            df2['pos'] = df2.index
            #df2.index = np.arange(df2.shape[0])
            df2['chrom'] = np.repeat(key, df2.shape[0])
            df2['posratio'] = df2['convs']/df2['covers']
            Outputdf = pd.concat([Outputdf,df2[order_columns]])
        return Outputdf.reset_index(drop=True)

    def createTag(self, d):
        return ''.join([''.join(key) + str(d[key]) + ';' for key in d.keys()])[:-1]

    def convInRead(self, read, qual=20):
        specific_conversions = {}
        total_content = {'a': 0, 'c': 0, 'g': 0, 't': 0}
        specific_conversions[('c', 'A')] = 0
        specific_conversions[('g', 'A')] = 0
        specific_conversions[('t', 'A')] = 0
        specific_conversions[('a', 'C')] = 0
        specific_conversions[('g', 'C')] = 0
        specific_conversions[('t', 'C')] = 0
        specific_conversions[('a', 'G')] = 0
        specific_conversions[('c', 'G')] = 0
        specific_conversions[('t', 'G')] = 0
        specific_conversions[('a', 'T')] = 0
        specific_conversions[('c', 'T')] = 0
        specific_conversions[('g', 'T')] = 0
        specific_conversions[('a', 'N')] = 0
        specific_conversions[('c', 'N')] = 0
        specific_conversions[('g', 'N')] = 0
        specific_conversions[('t', 'N')] = 0

        tC_loc = []
        aG_loc = []

        try:
            refseq = read.get_reference_sequence().lower()
        except (UnicodeDecodeError):
            refseq = ''
        except (AssertionError):
            return 0

        for base in total_content.keys():
            total_content[base] += refseq.count(base)
        for pair in read.get_aligned_pairs(with_seq=True):
            try:
                if pair[0] is not None and pair[1] is not None and pair[2] is not None:
                    if str(pair[2]).islower() and not read.query_qualities[pair[0]] < qual:
                        specific_conversions[(pair[2], read.seq[pair[0]])] += 1
                        if (pair[2], read.seq[pair[0]]) == ('t', 'C'):
                            tC_loc.append(pair[1])
                        if (pair[2], read.seq[pair[0]]) == ('a', 'G'):
                            aG_loc.append(pair[1])
            except (UnicodeDecodeError, KeyError):
                continue
        SC_tag = self.createTag(specific_conversions)
        TC_tag = self.createTag(total_content)

        if len(tC_loc) == 0:
            tC_loc.append(0)
        if len(aG_loc) == 0:
            aG_loc.append(0)
        return SC_tag, TC_tag, tC_loc, aG_loc


    @utils.add_log
    def addTags(self, bamfilename, outfile_bam, cellfile, strandednessfile, qual=20, thread=8):
        ContigLocs={}
        AnnoteLocs={}
        bamfile = pysam.AlignmentFile(bamfilename, 'rb')
        mod_bamfile = pysam.AlignmentFile(outfile_bam, mode='wb', template=bamfile)
        tmpread={}
        cellist = pd.read_csv(cellfile, header=None, index_col=0)
        cells = cellist.index
        strandedness = pd.read_csv(strandednessfile, header=None, index_col=0)
        genes=strandedness.index
        chrp=''

        class GeneError(Exception):
            pass

        for read in bamfile.fetch(until_eof=True):
            try:
                if not read.has_tag('GX'):
                    continue
                if read.get_tag("CB") not in cells:
                    continue
                if read.get_tag('GX') not in genes:
                    raise GeneError
                readid=read.get_tag("CB")+read.get_tag("UB")+read.get_tag("GX")
                chr1=read.reference_name

                if chr1!=chrp and chrp!='':
                    for x in tmpread:
                        mod_bamfile.write(tmpread[x][1])
                        ContigLocs, AnnoteLocs=self.CountConvperPos_read(tmpread[x][1], ContigLocs, AnnoteLocs)
                    tmpread={}

                tags = self.convInRead(read, qual)
                if tags==0: 
                    continue
                read.set_tag('SC', tags[0], 'Z')
                read.set_tag('TC', tags[1], 'Z')
                read.set_tag('TL', tags[2])
                read.set_tag('AL', tags[3])
                read.set_tag('ST', strandedness.loc[read.get_tag('XT')][1])
                
                if read.get_tag('ST') == '+':
                    stag = read.get_tag('TL')
                else:
                    stag = read.get_tag('AL')
                if len(stag) == 1 and stag[0] == 0:
                    tc = 0
                else:
                    tc = len(stag)
                if readid in tmpread:
                    if tc>tmpread[readid][0]:
                        tmpread[readid] = [tc,read]
                else:
                    tmpread[readid] = [tc,read]
                
            except (ValueError, KeyError):
                continue
            except (GeneError):
                print('{} is not in strand file, please check your files.'.format(read.get_tag("GX")))
                continue
            #except (Exception):
            #    print("convert error")
            #    sys.exit(1)
        
        for x in tmpread:
            mod_bamfile.write(tmpread[x][1])
            ContigLocs, AnnoteLocs=self.CountConvperPos_read(tmpread[x][1], ContigLocs, AnnoteLocs)

        bamfile.close()
        mod_bamfile.close()

        cmd = ['samtools sort -@', str(thread), '-o', outfile_bam+'.bam', outfile_bam]
        self.run_cmd(cmd)
        cmd = ['mv', outfile_bam+'.bam', outfile_bam]
        self.run_cmd(cmd)
        return ContigLocs,AnnoteLocs


@utils.add_log
def conversion(args):

    with Conversion(args) as runner:
        runner.run()


def get_opts_conversion(parser, sub_program):
    parser.add_argument('--strand', help='gene strand file, the format is "geneID,+/-"', required=True)
    parser.add_argument('--basequalilty', default=20, type=int ,help='min base quality of the read sequence', required=False)
    if sub_program:
        parser.add_argument(
            "--bam", help='featureCount bam(sortedByCoord), must have "MD" tag, set in star step', required=True)
        parser.add_argument("--cell", help='barcode cell list', required=True)
        parser = s_common(parser)
    return parser





import os
import subprocess

from celescope.tools import utils
from celescope.tools.mkref import Mkref, super_opts

class Mkref_snp(Mkref):
    """
    ## Features
    - Create dictionary file and fasta index for gatk SplitNCigarReads.
    (https://gatk.broadinstitute.org/hc/en-us/articles/360035531652-FASTA-Reference-genome-format) 
    Need to run `celescope rna mkref` first

    ## Output
    - fasta index
    - gatk dictionary file

    ## Usage
    ```
    # run celescope rna mkref first
    celescope snp mkref \\
     --genome_name Homo_sapiens_ensembl_99 \\
     --fasta Homo_sapiens.GRCh38.dna.primary_assembly.fa
    ```
    """


    @utils.add_log
    def build_fasta_index(self):
        cmd = (
            f'samtools faidx {self.fasta}'
        )
        self.build_fasta_index.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def build_fasta_dict(self):
        cmd = (
            f'gatk CreateSequenceDictionary -R {self.fasta}'
        )
        self.build_fasta_dict.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def run(self):
        assert os.path.exists(self.config_file), (
            f"{self.config_file} not exist! "
            "You need to build rna genome with 'celescope rna mkref' first."
        )
        super().run()
        self.build_fasta_index()
        self.build_fasta_dict()


def mkref(args):
    genome_type = 'snp'
    with Mkref_snp(genome_type, args, ) as runner:
        runner.run()


def get_opts_mkref(parser, sub_program):
    super_opts(parser, sub_program)




from celescope.tools.mkref import Mkref, super_opts

class Mkref_virus(Mkref):
    """
    ## Features
    - Create a virus genome reference directory.

    ## Output

    - STAR genome index files
    - Genome config file

    ## Usage
    ```
    celescope capture_virus mkref \\
        --genome_name EBV \\
        --fasta EBV_genome.fasta \\
    ```

    ```
    $ cat celescope_genome.config
    [genome]
    genome_type = virus
    fasta = EBV_genome.fasta
    genome_name = EBV
    ```
    """

    def run(self):
        super().run()
        self.build_star_index()

def mkref(args):
    genome_type = 'virus'
    with Mkref_virus(genome_type, args, ) as runner:
        runner.run()


def get_opts_mkref(parser, sub_program):
    super_opts(parser, sub_program)



import pandas as pd
import pysam
import os
import argparse
from xopen import xopen

UMI_10X_LEN = 10
TSO = "TTTCTTATATGGG"
SEQ_LEN = 150


def check_mkdir(dir_name):
    """if dir_name is not exist, make one"""
    if not os.path.exists(dir_name):
        os.system(f"mkdir -p {dir_name}")


def get_10X_barcode(cellranger_path):
    BARCODE_10X_FILE = os.path.dirname(cellranger_path) + "/lib/python/cellranger/barcodes/737K-august-2016.txt"
    return BARCODE_10X_FILE


def fastq_line(name, seq, qual):
    return f'@{name}\n{seq}\n+\n{qual}\n'


def convert_seq(sgr_barcode, umi, barcode_dict, barcodes_10X, seq2, qual2):
    '''
    barcode_dict - key:SGR barcode; value:10X barcode
    '''
    if sgr_barcode in barcode_dict:
        barcode_10X = barcode_dict[sgr_barcode]
    else:
        barcode_10X = barcodes_10X.readline().strip()
        barcode_dict[sgr_barcode] = barcode_10X

    if len(umi) > UMI_10X_LEN:
        umi_10X = umi[0:UMI_10X_LEN]
    elif len(umi) < UMI_10X_LEN:
        umi_10X = umi + 'C' * (UMI_10X_LEN - len(umi))
    else:
        umi_10X = umi

    seq2_insert = 90
    seq2_cut = 60

    new_seq2_1 = seq2[0:seq2_insert]
    new_seq2_2 = seq2[seq2_cut:] 

    new_seq1 = barcode_10X + umi_10X + TSO
    new_qual1 = 'J' * len(new_seq1)
    new_qual2_1 = qual2[0:len(new_seq2_1)]
    new_qual2_2 = qual2[seq2_cut:]

    return new_seq1, new_qual1, new_seq2_1, new_qual2_1, new_seq2_2, new_qual2_2


def run_convert(BARCODES_10X_FILE, fq2, outdir):

    sample = os.path.abspath(fq2).split('/')[-1]
    sample_prefix = '_'.join(sample.split('_')[:-1])
    barcodes_10X = open(BARCODES_10X_FILE, 'r')
    fq_file = pysam.FastxFile(fq2)

    out_fq1 = xopen(f'{outdir}/{sample_prefix}_S1_L001_R1_001.fastq.gz', 'w')
    out_fq2 = xopen(f'{outdir}/{sample_prefix}_S1_L001_R2_001.fastq.gz', 'w')   
    barcode_dict = {}

    for entry in fq_file:
        name = entry.name
        attrs = name.split('_')
        barcode = attrs[0]
        umi = attrs[1]
        seq = entry.sequence
        qual = entry.quality
        new_seq1, new_qual1, new_seq2_1, new_qual2_1, new_seq2_2, new_qual2_2 = convert_seq(barcode, umi, barcode_dict, barcodes_10X, seq, qual)
        
        out_fq1.write(fastq_line(name, new_seq1, new_qual1))
        out_fq2.write(fastq_line(name, seq, qual))

    out_fq1.close()
    out_fq2.close()
    barcodes_10X.close()

    barcode_record = pd.DataFrame()
    barcode_record['sgr'] = list(barcode_dict.keys())
    barcode_record['10X'] = [barcode_dict[i] for i in barcode_dict]
    barcode_record.to_csv(f'{outdir}/barcode_correspondence_file', sep='\t', index=False)

def main():
    parser = argparse.ArgumentParser(description='convert sgr fq to 10X')
    parser.add_argument('--soft_path', help='cellranger absolute path', required=True)
    parser.add_argument('--sgr_fq', help='read2 fastq file in 01.barcode', required=True)
    parser.add_argument('--outdir', help='output path of 10X format fastq file', required=True)
    args = parser.parse_args()

    check_mkdir(args.outdir)
    BARCODE_10X_FILE = get_10X_barcode(args.soft_path)
    run_convert(BARCODE_10X_FILE, args.sgr_fq, args.outdir)


if __name__ == '__main__':
    main()


from celescope.dynaseq.__init__ import __ASSAY__
from celescope.tools.multi import Multi
from celescope.tools.__init__ import FILTERED_MATRIX_DIR_SUFFIX, BARCODE_FILE_NAME


class Multi_dynaseq(Multi):

    """
    ## Usage

    ```
        multi_dynaseq\\
        --mapfile ./rna.mapfile\\
        --genomeDir /SGRNJ/Public/Database/genome/homo_mus\\
        --strand /SGRNJ03/Public/Database/genome/gene.strandedness.csv
    ```

    You need to generate strandness-file from gtf file. 
    The format is "geneID,strand", eg:
    ```
    ENSG00000223972,+
    ENSG00000227232,-
    ENSG00000278267,-
    ```
    """

    def star(self, sample):
        """
        """
        step = 'star'
        fq = f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq {fq} '
            f'--STAR_param "--outFilterScoreMinOverLread 0.3 --outFilterMatchNminOverLread 0.3 --outSAMattributes MD NH HI AS nM" '
        )
        self.process_cmd(cmd, step, sample, m=self.args.starMem, x=self.args.thread)

    def prep_map(self, sample):
        step = 'prep_map'
        arr = self.fq_dict[sample]
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq1 {arr[0]} --fq2 {arr[1]} '
            f'--STAR_param "--outFilterScoreMinOverLread 0.3 --outFilterMatchNminOverLread 0.3 --outSAMattributes MD NH HI AS nM"'
        )
        self.process_cmd(cmd, step, sample, m=self.args.starMem, x=self.args.starMem)

    def conversion(self, sample):
        step = 'conversion'
        bam = f'{self.outdir_dic[sample]["featureCounts"]}/{sample}_aligned_sortedByCoord_addTag.bam'
        cell = f'{self.outdir_dic[sample]["count"]}/{sample}_{FILTERED_MATRIX_DIR_SUFFIX[0]}/{BARCODE_FILE_NAME}'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--bam {bam} '
            f'--cell {cell} '
        )
        self.process_cmd(cmd, step, sample, m=8, x=1)

    def substitution(self, sample):
        step = 'substitution'
        bam = f'{self.outdir_dic[sample]["conversion"]}/{sample}.PosTag.bam'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--bam {bam} '
        )
        self.process_cmd(cmd, step, sample, m=1, x=1)

    def replacement(self, sample):
        step = 'replacement'
        bam = f'{self.outdir_dic[sample]["conversion"]}/{sample}.PosTag.bam'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--bam {bam} '
            f'--bg {self.col5_dict[sample]} '
        )
        self.process_cmd(cmd, step, sample, m=10, x=1)

    def replace_tsne(self, sample):
        step = 'replace_tsne'
        tsne_file = f'{self.outdir_dic[sample]["analysis"]}/{sample}_tsne_coord.tsv'
        mat_file = f'{self.outdir_dic[sample]["replacement"]}/{sample}.fraction_of_newRNA_matrix.txt'
        rep_file = f'{self.outdir_dic[sample]["replacement"]}/{sample}.fraction_of_newRNA_per_cell.txt'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--tsne {tsne_file} '
            f'--mat {mat_file} '
            f'--rep {rep_file} '
        )
        self.process_cmd(cmd, step, sample, m=1, x=1)


def main():
    multi = Multi_dynaseq(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()



import pandas as pd
import pysam

from collections import OrderedDict
from celescope.flv_trust4.__init__ import CHAIN, PAIRED_CHAIN
from celescope.flv_CR.summarize import Summarize
from celescope.tools import utils
from celescope.tools.step import s_common, Step
from celescope.tools.plotly_plot import Bar_plot


def gen_vj_annotation_metrics(df, seqtype):
    """
    Generate vdj Annotation Metrics from contig annotations file.
    """
    def get_vj_spanning_pair():
        """
        Get Productive V-J Spanning_Pair metric from annotation file
        Return productive chain pair number. eg: TRA/TRB or IGH/IGL, IGH/IGK.
        """
        df_productive = df[df['productive'] == True]
        
        if seqtype == "BCR":
            df_chain_heavy = df_productive[(df_productive['chain'] == 'IGH')] 
            df_chain_light = df_productive[(df_productive['chain'] == 'IGL') | (df_productive['chain'] =='IGK')]
        else:
            df_chain_heavy = df_productive[df_productive['chain'] == 'TRA']
            df_chain_light = df_productive[df_productive['chain'] == 'TRB']

        for _df in [df_chain_heavy, df_chain_light]:
            _df.drop_duplicates(['barcode'], inplace=True)

        VJ_Spanning_Pair_Cells = pd.merge(df_chain_heavy, df_chain_light, on='barcode', how='inner')
        
        return VJ_Spanning_Pair_Cells.shape[0]

    metric_dict = OrderedDict()
    chains, chain_pairs = CHAIN[seqtype], PAIRED_CHAIN[seqtype]

    metric_dict["Cells match with scRNA-seq analysis"] = len(set(df.barcode))
    metric_dict["Cells With Productive V-J Spanning Pair"] = get_vj_spanning_pair()

    for pair in chain_pairs:
        chain1, chain2 = pair.split('_')[0], pair.split('_')[1]
        cbs1 = set(df[(df['productive']==True)&(df['chain']==chain1)].barcode)
        cbs2 = set(df[(df['productive']==True)&(df['chain']==chain2)].barcode)
        metric_dict[f"Cells With Productive V-J Spanning ({chain1}, {chain2}) Pair"] = len(cbs1.intersection(cbs2))

    for chain in chains:
        metric_dict[f"Cells With {chain} Contig"] = len(set(df[df['chain']==chain].barcode))
        metric_dict[f"Cells With CDR3-annotated {chain} Contig"] = len(set(df[(df['chain']==chain)&(df['cdr3']!='None')].barcode))
        metric_dict[f"Cells With V-J Spanning {chain} Contig"] = len(set(df[(df['full_length']==True)&(df['chain']==chain)].barcode))
        metric_dict[f"Cells With Productive {chain} Contig"] =len(set(df[(df['productive']==True)&(df['chain'] == chain)].barcode))

    return metric_dict


def gen_clonotypes_table(df, out_clonotypes, seqtype):
    """
    Generate clonotypes.csv file
    """
    df = df[df['productive'] == True]
    df['chain_cdr3aa'] = df[['chain', 'cdr3']].apply(':'.join, axis=1)
    df = df.rename(columns={"chain_cdr3aa":"cdr3s_aa", "raw_clonotype_id":"clonotype_id"})
    df = df.dropna(subset=["clonotype_id"])
    df = df.sort_values("clonotype_id", key=lambda x: x.str.lstrip("clonotype").astype(int))
    
    sort_method = {"TCR": True, "BCR": False}
    cdr3_aa_dict = df.groupby("clonotype_id")["cdr3s_aa"].apply(set).to_dict()
    cdr3_aa_dict = {key: ';'.join(sorted(list(value), reverse=sort_method[seqtype])) for key, value in cdr3_aa_dict.items()}
    count_dict = df.groupby("clonotype_id")["barcode"].nunique().to_dict()
    
    df["frequency"] = df["clonotype_id"].apply(lambda x: count_dict[x])
    df["cdr3s_aa"] = df["clonotype_id"].apply(lambda x: cdr3_aa_dict[x])
    df = df.drop_duplicates("clonotype_id")
    df = df[["clonotype_id", "cdr3s_aa", "frequency"]]
    
    sum_frequency = df['frequency'].sum()
    df['proportion'] = df['frequency'].apply(lambda x: x/sum_frequency)

    df.to_csv(out_clonotypes, sep=',', index=False)


class Match(Step):
    """
    ## Features

    - Assembled results match with sc-RNA library.

    - Generate matched VDJ-annotation metrics, clonetypes table and bar-plot of clonetypes distribution in html.

    ## Output

    - `matched_contig_annotations.csv` High-level annotations of each high-confidence contigs from matched cell-associated barcodes.

    - `matched_contig.fasta` High-confidence contig sequences annotated in the matched_contig_annotations.csv.

    - `matched_productive_contig_annotations.csv` Annotations of each productive contigs from matched cell-associated barcodes. This is a subset of matched_contig_annotations.csv.

    - `matched_productive_contig.fasta` Productive contig sequences annotated in the matched_productive_contig_annotations.csv.

    - `clonotypes.csv` High-level descriptions of each clonotype where barcodes match with scRNA-Seq.

    """
    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.seqtype = args.seqtype
        self.chains = CHAIN[self.seqtype]
        self.pairs = PAIRED_CHAIN[self.seqtype]

        self.match_dir = args.match_dir
        if self.match_dir != 'None':
            self.match_cell_barcodes, _ = utils.get_barcode_from_match_dir(self.match_dir)

        self.filter_annotation = f'{args.summarize_out}/filtered_contig_annotations.csv'
        self.filter_fasta = f'{args.summarize_out}/filtered_contig.fasta'
        self.clonotypes = f'{args.summarize_out}/clonotypes.csv'

        # out
        self.match_annotation = f'{self.outdir}/matched_contig_annotations.csv'
        self.match_fasta = f'{self.outdir}/matched_contig.fasta'
        self.match_clonotypes = f'{self.outdir}/matched_clonotypes.csv'

    @utils.add_log
    def gen_matched_result(self):
        """
        Generate annotation and fasta files where barcodes matched with scRNA.
        """
        SGR_annotation_file = pd.read_csv(self.filter_annotation)
        match_annotation_file = SGR_annotation_file[SGR_annotation_file.barcode.isin(self.match_cell_barcodes)]
        match_annotation_file.to_csv(self.match_annotation, sep=',', index=False)

        SGR_fasta_file = pysam.FastxFile(self.filter_fasta)
        match_fasta_file = open(self.match_fasta, 'w')
        for entry in SGR_fasta_file:
            name = entry.name
            attrs = name.split('_')
            cb = attrs[0]
            if cb in self.match_cell_barcodes:
                new_name = cb + '_' + attrs[1] + '_' + attrs[2]
                seq = entry.sequence
                match_fasta_file.write(f'>{new_name}\n{seq}\n')
        match_fasta_file.close()

        Summarize.gen_productive_contig(match_annotation_file, self.match_fasta, self.outdir, prefix='matched_')
        
    @utils.add_log
    def gen_matched_clonotypes(self):
        """
        Generate clonotypes.csv file where barcodes match with scRNA
        """
        df_match = pd.read_csv(self.match_annotation)
        gen_clonotypes_table(df_match, self.match_clonotypes, self.seqtype)

    @utils.add_log
    def gen_matched_metrics(self):
        """
        Generate Matched vdj Annotation Metrics.
        """
        df_match = pd.read_csv(self.match_annotation)
        metrics_dict = gen_vj_annotation_metrics(df_match, self.seqtype)

        for k, v in metrics_dict.items():
            if k == "Cells match with scRNA-seq analysis":
                self.add_metric(
                    name = k,
                    value = v,
                    help_info="The intersection between VDJ cell barcodes and scRNA-Seq barcodes. All the following metrics are based on this intersection."
                )
            else:
                self.add_metric(
                    name = k,
                    value = v,
                    total = len(set(df_match.barcode))
                )

    @utils.add_log
    def gen_clonotypes_table(self):

        title = 'Clonetypes'
        raw_clonotypes = pd.read_csv(self.clonotypes, sep=',', index_col=None)
        raw_clonotypes['ClonotypeID'] = raw_clonotypes['clonotype_id'].apply(lambda x: x.strip('clonetype'))
        raw_clonotypes['Frequency'] = raw_clonotypes['frequency']
        raw_clonotypes['Proportion'] = raw_clonotypes['proportion'].apply(lambda x: f'{round(x*100, 2)}%')
        raw_clonotypes['CDR3_aa'] = raw_clonotypes['cdr3s_aa'].apply(lambda x: x.replace(';', '<br>'))

        table_dict = self.get_table_dict(
            title=title,
            table_id='clonetypes',
            df_table=raw_clonotypes[['ClonotypeID', 'CDR3_aa', 'Frequency', 'Proportion']]
        )
        self.add_data(table_dict=table_dict)

        raw_clonotypes['ClonotypeID'] = raw_clonotypes['ClonotypeID'].astype("int")
        raw_clonotypes.sort_values(by=['ClonotypeID'], inplace=True)
        Barplot = Bar_plot(df_bar=raw_clonotypes).get_plotly_div()
        self.add_data(Barplot=Barplot)

    @utils.add_log
    def run(self):
        if self.match_dir != 'None':
            self.gen_matched_result()
            self.gen_matched_clonotypes()
            self.gen_matched_metrics()
        self.gen_clonotypes_table()


def match(args):
    with Match(args, display_title="Match") as runner:
        runner.run()


def get_opts_match(parser, sub_program):
    parser.add_argument('--seqtype', help='TCR or BCR', choices=['TCR', 'BCR'], required=True)
    if sub_program:
        s_common(parser)
        parser.add_argument('--match_dir', help='scRNA-seq match directory', required=True)
        parser.add_argument('--summarize_out', help='assemble result in SGR barcode from summarize directory', required=True)
    return parser


import configparser
import subprocess
import sys
import os
import math

from celescope.tools import utils
from celescope.tools.__init__ import GENOME_CONFIG
from celescope.__init__ import HELP_DICT, __VERSION__


class Mkref():

    def __init__(self, genome_type, args, files=(), non_files=()):
        self.thread = args.thread
        self.genome_type = genome_type
        self.celescope_version = __VERSION__
        self.dry_run = args.dry_run
        self.STAR_param = args.STAR_param
        self.files = ('fasta',) + files
        self.non_files = ('genome_name',) + non_files

        for entry in self.files + self.non_files:
            value = str(getattr(args, entry))
            setattr(self, entry, value)

        self.config_dict = {}

        # get genomeSAindexNbases from fasta size
        fasta_size = os.path.getsize(self.fasta)
        self.genomeSAindexNbases = Mkref.get_SA(fasta_size)
        print(f'genomeSAindexNbases: {self.genomeSAindexNbases}')

        # out file
        self.config_file = GENOME_CONFIG

    @staticmethod
    def get_SA(fasta_size):
        """
        Get genomeSAindexNbases from fasta size. For small genomes, the parameter --genomeSAindexNbases must to be scaled down, with a typical value of min(14, log2(GenomeLength)/2 - 1). For example, for 1 megaBase genome, this is equal to 8, for 100 kiloBase genome, this is equal to 7.

        As mentioned in STAR manual, or 1 megaBase genome, this is equal to 9. But when the genome is relatively small, e.g. < 3kb,
        using large value may caused core dump error when mapping.

        Args:
            fasta_size: fasta size in bytes

        >>> Mkref.get_SA(10 ** 6)
        8
        >>> Mkref.get_SA(10 ** 5)
        7
        >>> Mkref.get_SA(10 ** 50)
        14
        """
        return min(int(math.log2(fasta_size) / 2 - 1), 14)

    def run(self):
        if self.dry_run:
            sys.exit(0)

    @utils.add_log
    def build_star_index(self):
        cmd = (
            f'STAR \\\n'
            f'--runMode genomeGenerate \\\n'
            f'--runThreadN {self.thread} \\\n'
            f'--genomeDir ./ \\\n'
            f'--genomeFastaFiles {self.fasta} \\\n'
            f'--genomeSAindexNbases {self.genomeSAindexNbases} \\\n'
        )
        if self.STAR_param:
            cmd += (" " + self.STAR_param)
        self.build_star_index.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def set_config_dict(self):
        """
        if some files are not in input arguments, set them in overwrite set_config_dict function
        """
        self.config_dict['genome_type'] = self.genome_type
        self.config_dict['celescope_version'] = self.celescope_version

        for entry in self.files + self.non_files:
            value = getattr(self, entry)
            print(f'{entry} : {value}')
            self.config_dict[entry] = value

    @utils.add_log
    def _write_config(self):
        config = configparser.ConfigParser()
        config['genome'] = self.config_dict

        with open(self.config_file, 'w') as config_handle:
            config.write(config_handle)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.set_config_dict()
        self._write_config()

    @staticmethod
    def get_file_path(raw_file_path, genomeDir):
        """
        if raw_file_path is not absolute path and not None,
        add genomeDir
        if 'NONE'(str), return None

        >>> raw_file_path = '/root/Homo_sapiens.GRCh38.92.chr.gtf'
        >>> Mkref.get_file_path(raw_file_path, 'fake')
        '/root/Homo_sapiens.GRCh38.92.chr.gtf'

        >>> raw_file_path = 'None'
        >>> Mkref.get_file_path(raw_file_path, 'fake')
        """
        if (not raw_file_path) or (raw_file_path == 'None'):
            return None

        file_path = raw_file_path
        if not file_path.startswith('/'):
            file_path = f'{genomeDir}/{file_path}'

        return file_path            

    @staticmethod
    def parse_genomeDir(genomeDir, files=()):
        """
        Parse genomeDir and return a dict.
        """
        files = ('fasta',) + files

        config_file = f'{genomeDir}/{GENOME_CONFIG}'
        config = configparser.ConfigParser()
        config.read_file(open(config_file))
        genome = dict(config['genome'])

        for entry in files:
            if entry not in genome:
                raise ValueError(f'{entry} not in {config_file}')
            genome[entry] = Mkref.get_file_path(genome[entry], genomeDir)        

        return genome


def super_opts(parser, sub_program):
    if sub_program:
        parser.add_argument("--thread", help="Default=6. Threads to use.", default=6)
        parser.add_argument("--genome_name", help="Required, genome name. ", required=True)
        parser.add_argument("--dry_run", help="Only write config file and exit.", action='store_true')
        parser.add_argument("--fasta", help=HELP_DICT['fasta'], required=True)
        parser.add_argument('--STAR_param', help=HELP_DICT['additional_param'], default="")



STEPS = [
    'mkref',
    'sample',
    'barcode',
    'cutadapt',
    'star',
    'prep_map',
    "featureCounts",
    "count",
    'analysis']
__ASSAY__ = 'rna'
REMOVE_FROM_MULTI = {
    'mkref',
    'barcode',
    'cutadapt',
    'star',
}





"""
from celescope.tools import utils
from celescope.tools.step import Step, s_common
import numpy as np
import pandas as pd
# from bs4 import BeautifulSoup

'''
    # def check(self, sample):
    #     step = 'check'
    #     cmd_line = self.get_cmd_line(step, sample)
    #     rep = f'{self.outdir_dic[sample]["assemble"]}/stat.txt'
    #     rep_bc = f'{self.outdir_dic[sample]["barcode"]}/stat.txt'
    #     html = f'{sample}/{sample}_report.html'
    #     cmd = (
    #         f'{cmd_line} '
    #         f'--rep {rep} '
    #         f'--rep_bc {rep_bc} '
    #         f'--html {html}'
    #     )
    #     self.process_cmd(cmd, step, sample, m=5, x=1)
    
The warning metrics of samples' analysis result are marked in red in the report
'''

warnings = {
    'Q30 of Barcodes':0.90,
    'Q30 of UMIs':0.90,
    'Estimated Number of Cells':10.0,
    'Reads Mapped to Any V(D)J Gene':0.60,
    'Cells With Productive V-J Spanning Pair':0.30,
    'Median used TRA UMIs per Cell':0.1,
    'Median used TRB UMIs per Cell':0.1,
    'Median used IGH UMIs per Cell':0.1,
    'Median used IGL UMIs per Cell':0.1,
    'Median used IGK UMIs per Cell':0.1  
           }

warnings_help = {
    'Q30 of Barcodes': "Ideal > 90%. Application performance may be affected.",
    'Q30 of UMIs': "Ideal > 90%. Application performance may be affected.",
    'Estimated Number of Cells': "Ideal >= 10. This usually indicates poor cell quality, poor library quality, or poor sequencing quality. Application performance is likely to be affected.",
    'Reads Mapped to Any V(D)J Gene': "Ideal > 60%. This can indicate poor specificity of the V(D)J enrichment, use of the wrong germline reference, or the use of an unsupported chemistry type (e.g., using Single Cell 3\' for V(D)J assembly). Application performance may be affected.",
    'Cells With Productive V-J Spanning Pair': "This can indicate poor cell quality, low yield from the RT reaction, poor specificity of the V(D)J enrichment, poor sequencing quality, or the use of an unsupported chemistry type (e.g., using Single Cell 3\' for V(D)J assembly). Application performance may be affected.",
    'Median used TRA UMIs per Cell': "Ideal > 0. This can indicate cells with extremely low TRA expression, poor cell quality, low yield from the RT reaction, or the use of an unsupported chemistry type (e.g., using Single Cell 3\' for V(D)J assembly). Application performance may be affected.",
    'Median used TRB UMIs per Cell': "Ideal > 0. This can indicate cells with extremely low TRB expression, poor cell quality, low yield from the RT reaction, or the use of an unsupported chemistry type (e.g., using Single Cell 3\' for V(D)J assembly). Application performance may be affected.",
    'Median used IGH UMIs per Cell': "Ideal > 0. This can indicate cells with extremely low IGH expression, poor cell quality, low yield from the RT reaction, or the use of an unsupported chemistry type (e.g., using Single Cell 3\' for V(D)J assembly). Application performance may be affected.",
    'Median used IGL UMIs per Cell': "Ideal > 0. This can indicate cells with extremely low IGL expression, poor cell quality, low yield from the RT reaction, or the use of an unsupported chemistry type (e.g., using Single Cell 3\' for V(D)J assembly). Application performance may be affected.",
    'Median used IGK UMIs per Cell': "Ideal > 0. This can indicate cells with extremely low IGK expression, poor cell quality, low yield from the RT reaction, or the use of an unsupported chemistry type (e.g., using Single Cell 3\' for V(D)J assembly). Application performance may be affected."
           }


class Check(Step):
    
    def __init__(self, args, step_name):
        Step.__init__(self, args, step_name)
    
    # input
        self.rep_bc = args.rep_bc
        self.rep = args.rep
        self.seqtype = args.seqtype
        self.html = args.html
        
        self.warning_list=[]


        if self.seqtype =='TCR':
            self.target = [
                'Q30 of Barcodes','Q30 of UMIs','Estimated Number of Cells','Reads Mapped to Any V(D)J Gene','Cells With Productive V-J Spanning Pair',
                'Median used TRA UMIs per Cell','Median used TRB UMIs per Cell']
        elif self.seqtype == 'BCR':
            self.target = [
                'Q30 of Barcodes','Q30 of UMIs','Estimated Number of Cells','Reads Mapped to Any V(D)J Gene','Cells With Productive V-J Spanning Pair',
                'Median used IGH UMIs per Cell','Median used IGL UMIs per Cell','Median used IGK UMIs per Cell']


    @utils.add_log
    def run_check(self):
        rep = dict()
        warning_summary = []

        with open(self.rep_bc,'r') as f:
            lines = (line.strip().split(':') for line in f)
            for line in lines:
                rep[line[0]] = line[1]
        with open(self.rep,'r') as f:
            lines = (line.strip().split(':') for line in f)
            for line in lines:
                rep[line[0]] = line[1]
        
        rep1 = {key:value for key,value in rep.items() if key in self.target}

        for i in rep1.keys():
            rep1[i] = rep1[i].strip()
            if ',' in rep1[i]:
                rep1[i] = rep1[i].replace(',','')
            if '(' in rep1[i]:
                rep1[i] = rep1[i].split('(')[1].strip(')')
            if '%' in rep1[i]:
                rep1[i] = round(float(rep1[i].strip('%'))/100,2)
            else:
                rep1[i] = float(rep1[i])

        for key in rep1:
            if rep1[key] <= warnings[key]:
                warning_summary.append({
                    'item': key,
                    'count': rep1[key],
                    'total_count': np.nan
                })
                
                self.warning_list.append(key)

        if len(self.warning_list) > 0:
            stat_file = self.outdir + '/stat.txt'
            sum_df = pd.DataFrame(warning_summary, columns=['item', 'count', 'total_count'])
            utils.gen_stat(sum_df, stat_file)

    @utils.add_log
    def check_html(self):

        with open(self.html,'r',encoding='utf-8') as file:
            pcontent = file.read()
        sp = BeautifulSoup(pcontent,'html.parser')

        for i in sp.find_all('td'):
            if i.string in self.warning_list:
                new_tag = sp.new_tag("font", color = 'red')
                new_tag.string = i.string
                i.string.replace_with(new_tag)
                new_tag1 = sp.new_tag("font", color = 'red')
                new_tag1.string = i.next_sibling.next_sibling.string
                i.next_sibling.next_sibling.string.replace_with(new_tag1)

        for i in sp.find_all('p'):
            try:
                if i.b.string.strip().strip(':') in self.warning_list:
                    new_tag = sp.new_tag("font", color = 'red')
                    new_tag.string = i.b.string
                    i.b.string.replace_with(new_tag)
                    new_tag1 = sp.new_tag("font", color = 'black')
                    new_tag1.string =  warnings_help[i.b.string.strip().strip(':')]
                    i.b.next_sibling.string.replace_with(new_tag1)
            except StopIteration:
                pass
        
        with open(f'{self.outdir}/{self.sample}_report.html','w',encoding='utf-8') as fp:
            fp.write(sp.prettify())

    def run(self):    
        self.run_check()
        if len(self.warning_list)>0:
            self.check_html()

def check(args):
    step_name = 'check'
    check_obj = Check(args, step_name)
    check_obj.run()

def get_opts_check(parser, sub_program):
    parser.add_argument('--seqtype', help='TCR or BCR', choices=['TCR', 'BCR'], required=True)

    if sub_program:
        parser = s_common(parser)
        parser.add_argument('--rep', help = 'assemble report file', required = True)
        parser.add_argument('--rep_bc', help = 'barcode report file', required = True)
        parser.add_argument('--html', help = 'html report file', required = True)

"""


STEPS = [
    'sample',
    'barcode',
    'cutadapt',
    'star',
    "star_virus",
    "featureCounts",
    "count",
    "count_virus",
    'analysis_rna_virus',
]
__ASSAY__ = 'rna_virus'
IMPORT_DICT = {
    'star': 'celescope.rna'
}



import numpy as np
import pandas as pd
import scipy.io
import scipy.sparse
import glob
import scanpy as sc
import anndata as ad

from celescope.tools import utils
from celescope.tools.step import Step, s_common
from celescope.__init__ import HELP_DICT
from celescope.tools.matrix import CountMatrix, Features


TAG_COL = 'tag_name'


def filter_fun(df):
    """
    Filter data according to mean and standard deviation
    """
    df['filter_num'] = np.mean(df,axis=1)+np.std(df,axis=1)
    for index,_ in df.iterrows():
        df.loc[index] = df.loc[index].mask(df.loc[index]<df.loc[index]['filter_num'],0)
    df.drop(columns='filter_num',inplace=True)
    return df



class CountMatrix2(CountMatrix):
    @classmethod
    def from_dataframe2(cls, df,features: Features, barcodes=None):
        matrix = df
        mtx = scipy.sparse.coo_matrix(matrix, matrix.shape, int)
        return cls(features,barcodes,mtx)


class Count_cite(Step):

    def __init__(self, args, display_title):
        super().__init__(args, display_title)

        self.df_read_count = pd.read_csv(args.read_count_file, sep='\t', index_col=0)

        self.match_dict = utils.parse_match_dir(args.match_dir)
        self.match_barcode = self.match_dict['match_barcode']
        self.match_matrix_dir = self.match_dict['matrix_dir']

        # input
        self.tsne_coord = glob.glob(f'{args.match_dir}/*analysis*/*tsne_coord.tsv')[0]

        # out
        self.mtx = f'{self.out_prefix}_citeseq.mtx.gz'
        self.raw_matrix_dir = f'{self.out_prefix}_raw_citeseq_matrix'
        self.filtered_matrix_dir  = f'{self.out_prefix}_filtered_citeseq_matrix'
        self.filtered_tsne_coord = f'{self.out_prefix}_filtered_tsne_coord.tsv'
        

    @utils.add_log
    def run(self):
        mapped_read = int(self.df_read_count['read_count'].sum())

        # in cell
        df_read_count_in_cell = self.df_read_count[self.df_read_count.index.isin(self.match_barcode)]
        mapped_read_in_cell = int(df_read_count_in_cell['read_count'].sum())
        self.add_metric(
            name='Mapped Reads in Cells',
            value=mapped_read_in_cell,
            total=mapped_read,
        )

        tag_names = df_read_count_in_cell[TAG_COL].unique()
        features_raw = Features(tag_names)
        
        # raw_matrix
        raw_citeseq_matrix = CountMatrix.from_dataframe(df_read_count_in_cell, features_raw, barcodes=self.match_barcode, row=TAG_COL, column='barcode', value='UMI')
        rna_matrix = CountMatrix.from_matrix_dir(matrix_dir=self.match_matrix_dir)
        raw_merged_matrix = rna_matrix.concat_by_barcodes(raw_citeseq_matrix)
        raw_merged_matrix.to_matrix_dir(self.raw_matrix_dir)

        # UMI
        df_UMI_in_cell = df_read_count_in_cell.reset_index().groupby([
            'barcode', TAG_COL]).agg({'UMI': 'count'})

        df_UMI_in_cell = df_UMI_in_cell.reset_index()
        df_UMI_in_cell = df_UMI_in_cell.pivot(
            index='barcode', columns=TAG_COL, values='UMI')
        df_cell = pd.DataFrame(index=self.match_barcode)
        df_UMI_cell = pd.merge(
            df_cell,
            df_UMI_in_cell,
            how="left",
            left_index=True,
            right_index=True)

        # fillna
        df_UMI_cell.fillna(0, inplace=True)
        df_UMI_cell = df_UMI_cell.astype(int)
        df_UMI_cell_out = df_UMI_cell.T
        df_UMI_cell_out.to_csv(self.mtx, sep='\t', compression='gzip')

        
        # fix
        ## 1.normalize;2.filter
        ### normalize
        obs = pd.DataFrame(index=df_UMI_cell_out.columns)
        var = pd.DataFrame(df_UMI_cell_out.index,index=df_UMI_cell_out.index,columns=['ADT'])
        mdata_citeseq = ad.AnnData(np.array(df_UMI_cell_out.T),obs=obs,var=var)
        sc.pp.normalize_total(
                mdata_citeseq,
                target_sum=1e4,
                inplace=True,
            )
        sc.pp.log1p(mdata_citeseq)
        ### filter
        df_filtered = filter_fun(mdata_citeseq.to_df().T)
        features_filtered = Features(df_filtered.index.to_list())
        filtered_citeseq_matrix = CountMatrix2.from_dataframe2(df_filtered, features_filtered, barcodes=self.match_barcode)
        filtered_merged_matrix = rna_matrix.concat_by_barcodes(filtered_citeseq_matrix)
        filtered_merged_matrix.to_matrix_dir(self.filtered_matrix_dir)
        
        
        # filtered_tsne.csv
        df_tsne = pd.read_csv(self.tsne_coord,sep="\t")
        if 'Unnamed: 0' in df_tsne.columns:
            df_tsne.rename(columns={'Unnamed: 0': 'barcode'}, inplace=True)
            df_tsne = df_tsne.set_index('barcode')

        df_citeseq = mdata_citeseq.to_df()
        df_tsne = pd.concat([df_tsne,df_citeseq],axis=1)
        df_tsne.fillna(0,inplace=True)
        df_tsne.to_csv(self.filtered_tsne_coord,sep='\t')


        # UMI
        UMIs = df_UMI_cell.apply(sum, axis=1)
        median_umi = round(np.median(UMIs), 2)
        mean_umi = round(np.mean(UMIs), 2)
        self.add_metric(
            name='Median UMI per Cell',
            value=float(median_umi),
        )
        self.add_metric(
            name='Mean UMI per Cell',
            value=float(mean_umi),
        )


def get_opts_count_cite(parser, sub_program):
    if sub_program:
        parser.add_argument("--match_dir", help=HELP_DICT['match_dir'], required=True)
        parser.add_argument("--read_count_file", help="tag read count file")
        s_common(parser)


def count_cite(args):

    with Count_cite(args, display_title='Count') as runner:
        runner.run()



import json
import pandas as pd
import pysam
import os 
import subprocess

from celescope.tools import utils
from celescope.tools.step import s_common, Step


class Summarize(Step):
    """
    ## Features

    - Convert 10X barcode of assemble result back to SGR barcode.

    - Generate Productive contigs sequences and annotation files.

    - Generate VDJ-annotation metrics in html.

    ## Output

    - `filtered_contig_annotations.csv` High-level annotations of each high-confidence contigs from cell-associated barcodes.

    - `filtered_contig.fasta` High-confidence contig sequences annotated in the filtered_contig_annotations.csv.

    - `productive_contig_annotations.csv` Annotations of each productive contigs from cell-associated barcodes. This is a subset of filtered_contig_annotations.csv.

    - `productive_contig.fasta` Productive contig sequences annotated in the productive_contig_annotations.csv.

    - `clonotypes.csv` High-level descriptions of each clonotype.

    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.seqtype = args.seqtype
        self.version = os.path.dirname(args.soft_path).split('/')[-1].split('-')[-1]

        with open(args.barcode_convert_json, 'r') as f:
            self.tenX_sgr = json.load(f)

        # assemble result directory in 10X barcode.
        if args.not_refine:
            self.assemble_out = args.assemble_out
        else:
            self.assemble_out = args.assemble_out.replace('/outs', '_refine/outs')

        # in files
        annotation_file = f'{self.assemble_out}/filtered_contig_annotations.csv'
        self.df_annotation = pd.read_csv(annotation_file, sep=',', index_col=None)
        tenX_fasta_file = f'{self.assemble_out}/filtered_contig.fasta'
        self.tenX_fasta_fh = pysam.FastxFile(tenX_fasta_file)

        # out
        self.filtered_contig_annotations = f'{self.outdir}/filtered_contig_annotations.csv'
        self.filter_contig_fasta = f'{self.outdir}/filtered_contig.fasta'
    
    @utils.add_log
    def convert_barcode_to_SGR(self):
        """
        Convert 10X barcode to SGR barcode format.
        """

        self.df_annotation['barcode'] = self.df_annotation['barcode'].apply(lambda x: self.tenX_sgr[x.split('-')[0]])
        self.df_annotation['contig_id'] = self.df_annotation['contig_id'].apply(lambda x: 
            self.tenX_sgr[x.split('-')[0]]+'_'+x.split('_')[1]+'_'+x.split('_')[2])
        if int(self.version.split('.')[0]) < 4:
            self.df_annotation['productive'].replace({'True': True, 'None': False}, inplace=True)

        self.df_annotation.to_csv(self.filtered_contig_annotations, sep=',', index=False)
    

        with open(self.filter_contig_fasta, 'w') as f:
            for entry in self.tenX_fasta_fh:
                name = entry.name
                seq = entry.sequence
                attrs = name.split('_')
                new_name = self.tenX_sgr[attrs[0].split('-')[0]] + '_' + attrs[1] + '_' + attrs[2]
                f.write(f'>{new_name}\n{seq}\n')
        
        cmd = f'cp {self.assemble_out}/clonotypes.csv {self.outdir}'
        subprocess.check_call(cmd, shell=True)

    @staticmethod
    @utils.add_log
    def gen_productive_contig(annotation_file, fasta_file, outdir, prefix=''):
        """
        Generate productive_contig_annotation.csv and productive_contig.fasta file.
        """
        productive_contig = annotation_file[annotation_file['productive'] == True]
        productive_contig_id = set(productive_contig['contig_id'])
        productive_contig.to_csv(f'{outdir}/{prefix}productive_contig_annotations.csv', sep=',', index=False)

        productive_fasta = open(f'{outdir}/{prefix}productive_contig.fasta', 'w')
        with pysam.FastxFile(fasta_file, 'r') as f:
            for read in f:
                if read.name in productive_contig_id:
                    productive_fasta.write(">" + read.name + "\n" + read.sequence + "\n")
        productive_fasta.close()

    def run(self):
        self.convert_barcode_to_SGR()
        Summarize.gen_productive_contig(self.df_annotation, self.filter_contig_fasta, self.outdir)


def summarize(args):
    with Summarize(args) as runner:
        runner.run()


def get_opts_summarize(parser, sub_program):
    parser.add_argument('--seqtype', help='TCR or BCR', choices=['TCR', 'BCR'], required=True)
    parser.add_argument('--soft_path', help='soft path for cellranger')
    parser.add_argument("--not_refine", help="Do not perform the refine step. ", action="store_true")
    if sub_program:
        s_common(parser)
        parser.add_argument('--barcode_convert_json', help='json file', required=True)
        parser.add_argument('--assemble_out', help='directory of cellranger assemble result', required=True)
    return parser


"""
count step
"""

import os
import random
import sys

import numpy as np
import pandas as pd

from celescope.tools import utils
from celescope.__init__ import HELP_DICT
from celescope.tools.__init__ import (FILTERED_MATRIX_DIR_SUFFIX, RAW_MATRIX_DIR_SUFFIX)
from celescope.tools.emptydrop_cr import get_plot_elements
from celescope.tools.emptydrop_cr.cell_calling_3 import cell_calling_3
from celescope.tools.step import Step, s_common
from celescope.rna.mkref import Mkref_rna
from celescope.tools.plotly_plot import Line_plot
from celescope.tools.matrix import CountMatrix
from celescope.tools import reference

TOOLS_DIR = os.path.dirname(__file__)
random.seed(0)
np.random.seed(0)

# downsample.csv
READ_FRACTION = 'read_fraction'
MEDIAN_GENE_NUMBER = 'median_gene_number'
SATURATION = 'saturation'

# Plot axis title in HTML
X_TITLE = 'Read Fraction'
SATURATION_Y_TITLE = 'Sequencing Saturation(%)'
MEDIAN_GENE_Y_TITLE = 'Median Genes per Cell'


class Count(Step):
    """
    ## Features
    - Cell-calling: Distinguish cell barcodes from background barcodes. 
    - Generate expression matrix.
    ## Output
    - `{sample}_raw_feature_bc_matrix` The expression matrix of all detected barcodes in [Matrix Market Exchange Formats](
        https://math.nist.gov/MatrixMarket/formats.html). 
    - `{sample}_filtered_feature_bc_matrix` The expression matrix of cell barcodes in Matrix Market Exchange Formats. 
    - `{sample}_count_detail.txt.gz` 4 columns: 
        - barcode  
        - gene ID  
        - UMI count  
        - read_count  
    - `{sample}_counts.txt` 6 columns:
        - Barcode: barcode sequence
        - read: read count of each barcode
        - UMI: UMI count for each barcode
        - geneID: gene count for each barcode
        - mark: cell barcode or backgound barcode.
            `CB` cell  
            `UB` background  
    - `{sample}_downsample.tsv` Subset a fraction of reads and calculate median gene number and sequencing saturation.
    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)
        self.force_cell_num = args.force_cell_num
        self.cell_calling_method = args.cell_calling_method
        self.expected_cell_num = int(args.expected_cell_num)

        # set
        gtf_file = Mkref_rna.parse_genomeDir(args.genomeDir)['gtf']
        gp = reference.GtfParser(gtf_file)
        gp.get_id_name()
        self.features = gp.get_features()
        self.downsample_dict = {}

        # output files
        self.marked_count_file = f'{self.outdir}/{self.sample}_counts.txt'
        self.raw_matrix_dir = f'{self.outdir}/{self.sample}_{RAW_MATRIX_DIR_SUFFIX[0]}'
        self.cell_matrix_dir = f'{self.outdir}/{self.sample}_{FILTERED_MATRIX_DIR_SUFFIX[0]}'
        self.downsample_file = f'{self.outdir}/{self.sample}_downsample.tsv'

    def get_df_line(self):
        df_line = pd.read_csv(self.downsample_file, sep="\t", header=0)
        df_line.rename(columns={
            READ_FRACTION: X_TITLE,
            MEDIAN_GENE_NUMBER: MEDIAN_GENE_Y_TITLE,
            SATURATION: SATURATION_Y_TITLE,
        }, inplace=True)

        return df_line

    @staticmethod
    def get_df_cell(df, cell_bc):
        return df.loc[df['Barcode'].isin(cell_bc), :]

    @utils.add_log
    def run(self):
        df = pd.read_table(self.args.count_detail, header=0, dtype={'geneID':str})

        # write all matrix
        self.write_sparse_matrix(df, self.raw_matrix_dir)
        df_sum = self.get_df_sum(df)
        total_reads = sum(df_sum['read'])
        # call cells
        cell_bc, _threshold = self.cell_calling(df_sum)
        # write marked_df_sum
        self.write_marked_df_sum(df_sum, cell_bc)
        # export cell matrix
        df_cell = self.get_df_cell(df, cell_bc)
        self.write_sparse_matrix(df_cell, self.cell_matrix_dir)
        total_cell_gene = len(np.unique(df_cell['geneID']))
        del df
        df_sum_cell = self.get_df_sum(df_cell)
        del df_sum
        # downsample
        downsample_dict = self.downsample(df_cell)
        # metrics
        self.add_count_metrics(total_reads, total_cell_gene, cell_bc, df_sum_cell, downsample_dict)
        # plot
        self.add_plot_data()

    def add_count_metrics(self, total_reads, total_cell_gene, cell_bc, df_sum_cell, downsample_dict):
        n_cells = len(cell_bc)
        self.add_metric(
            name='Estimated Number of Cells',
            value=n_cells,
            help_info=f'the number of barcodes considered as cell-associated. The cell calling method used for this sample is {self.cell_calling_method}.'
        )

        cell_reads = sum(df_sum_cell['read'])
        fraction_reads_in_cells = round(float(cell_reads) / total_reads * 100, 2)
        self.add_metric(
            name='Fraction Reads in Cells',
            value=fraction_reads_in_cells,
            display=f'{fraction_reads_in_cells}%',
            help_info='the fraction of uniquely-mapped-to-transcriptome reads with cell-associated barcodes'
        )

        try:
            valid_read_number = self.get_slot_key(
                slot='metrics',
                step_name='barcode',
                key='Valid Reads',
            )
        except KeyError:
            self.add_count_metrics.logger.warning('Will not output `Mean Reads per Cell`')
        else:
            mean_reads_per_cell = int(valid_read_number / n_cells)
            self.add_metric(
                name='Mean Reads per Cell',
                value=mean_reads_per_cell,
                help_info='the number of valid reads divided by the estimated number of cells'
            )

        median_umi_per_cell = int(np.median(df_sum_cell['UMI']))
        self.add_metric(
            name='Median UMI per Cell',
            value=median_umi_per_cell,
            help_info='the median number of UMI counts per cell-associated barcode'
        )

        self.add_metric(
            name='Total Genes',
            value=total_cell_gene,
            help_info='the number of genes with at least one UMI count in any cell'
        )

        self.add_metric(
            name='Median Genes per Cell',
            value=downsample_dict[MEDIAN_GENE_NUMBER][-1],
            help_info='the median number of genes detected per cell-associated barcode'
        )

        saturation = downsample_dict[SATURATION][-1]
        self.add_metric(
            name='Saturation',
            value=saturation,
            display=f'{saturation}%',
            help_info='the fraction of UMI originating from an already-observed UMI. '
        )


    def add_plot_data(self):

        df_line = self.get_df_line()

        line_saturation = Line_plot(df_line=df_line,title="Sequencing Saturation",x_title=X_TITLE,
                                    y_title=SATURATION_Y_TITLE, y_range=[0,100], section=False).get_plotly_div()
        self.add_data(line_saturation=line_saturation)
        line_median = Line_plot(df_line=df_line,title="Median Genes per Cell", x_title=X_TITLE,
                                y_title=MEDIAN_GENE_Y_TITLE).get_plotly_div()
        self.add_data(line_median=line_median)

        self.add_data(chart=get_plot_elements.plot_barcode_rank(self.marked_count_file))


    
    @utils.add_log
    def cell_calling(self, df_sum):
        """
        Returns:
            cell_bc: set
            UMI_threshold: int
        """
        cell_calling_method = self.cell_calling_method

        if (self.force_cell_num is not None) and (self.force_cell_num != 'None'):
            cell_bc, UMI_threshold = self.force_cell(df_sum)
        elif cell_calling_method == 'auto':
            cell_bc, UMI_threshold = self.auto_cell(df_sum)
        elif cell_calling_method == 'EmptyDrops_CR':
            cell_bc, UMI_threshold = self.emptydrop_cr_cell(df_sum)
        cell_bc = set(cell_bc)
        return cell_bc, UMI_threshold

    @utils.add_log
    def force_cell(self, df_sum):
        force_cell_num = int(self.force_cell_num)

        df_barcode_count = df_sum.groupby(
            ['UMI']).size().reset_index(
            name='barcode_counts')
        sorted_df = df_barcode_count.sort_values("UMI", ascending=False)
        sorted_df["barcode_cumsum"] = sorted_df["barcode_counts"].cumsum()
        num_points = sorted_df.shape[0]
        for i in range(num_points):
            if sorted_df.iloc[i, :]["barcode_cumsum"] >= force_cell_num:
                index_low = i - 1
                index_high = i
                break
        df_sub = sorted_df.iloc[index_low: index_high + 1, :]
        distance = abs(df_sub["barcode_cumsum"] - force_cell_num)
        actual_index = np.argmin(distance)
        threshold = df_sub.iloc[actual_index, :]['UMI']
        cell_bc = Count.get_cell_bc(df_sum, threshold, col='UMI')
        return cell_bc, threshold

    @staticmethod
    def find_threshold(df_sum, idx):
        return int(df_sum.iloc[idx - 1, df_sum.columns == 'UMI'])

    @staticmethod
    def get_cell_bc(df_sum, threshold, col='UMI'):
        return list(df_sum[df_sum[col] >= threshold].index)

    @utils.add_log
    def auto_cell(self, df_sum):
        idx = int(self.expected_cell_num * 0.01)
        barcode_number = df_sum.shape[0]
        idx = int(min(barcode_number, idx))
        if idx == 0:
            sys.exit("cell number equals zero!")
        # calculate read counts threshold
        threshold = int(Count.find_threshold(df_sum, idx) * 0.1)
        threshold = max(1, threshold)
        cell_bc = Count.get_cell_bc(df_sum, threshold)

        return cell_bc, threshold

    @utils.add_log
    def emptydrop_cr_cell(self, df_sum):
        cell_bc, initial_cell_num = cell_calling_3(self.raw_matrix_dir, self.expected_cell_num)
        threshold = Count.find_threshold(df_sum, initial_cell_num)
        return cell_bc, threshold

    @staticmethod
    def get_df_sum(df, sort_by='UMI'):
        '''
        Returns:
            df_sum: 3 cols
                read: int
                UMI: int
                geneID: str
        '''

        df_sum = df.groupby('Barcode').agg({
            'read': 'sum',
            'UMI': 'count',
            'geneID': 'nunique'
        })
        df_sum = df_sum.sort_values(sort_by, ascending=False)
        return df_sum

    def write_marked_df_sum(self, df_sum, cell_bc):
        df_sum.loc[:, 'mark'] = 'UB'
        df_sum.loc[df_sum.index.isin(cell_bc), 'mark'] = 'CB'
        df_sum.to_csv(self.marked_count_file, sep='\t')

    @utils.add_log
    def write_sparse_matrix(self, df, matrix_dir):

        count_matrix = CountMatrix.from_dataframe(df, self.features, value="UMI")
        count_matrix.to_matrix_dir(matrix_dir)

    @utils.add_log
    def cell_summary(self, df, cell_bc):

        df.loc[:, 'mark'] = 'UB'
        df.loc[df['Barcode'].isin(cell_bc), 'mark'] = 'CB'
        CB_total_Genes = df.loc[df['mark'] == 'CB', 'geneID'].nunique()
        CB_reads_count = df.loc[df['mark'] == 'CB', 'count'].sum()
        reads_mapped_to_transcriptome = df['count'].sum()
        return(CB_total_Genes, CB_reads_count, reads_mapped_to_transcriptome)

    @staticmethod
    def sub_sample(fraction, df_cell, new_df, cell_read_index):
        """
        saturation = 1 - n_deduped_reads / n_reads

        n_deduped_reads = Number of unique (valid cell-barcode, valid UMI, gene) combinations among confidently mapped reads. If a UMI has two reads mapped to one gene, but mapped to different locations of the gene, these two reads are
        considered unique.
        n_reads = Total number of (confidently mapped, valid cell-barcode, valid UMI) reads.
        Args:
            fration: subsmaple fration
            df_cell: in cell df with (Barcode geneID UMI count) 
            new_df: two columns: [original df_cell index, duplicate count]
            cell_read_index: df_cell repeat index
        """

        cell_read = new_df['dcount'].sum()
        frac_n_read = int(cell_read * fraction)
        subsample_read_index = cell_read_index[:frac_n_read]
        index_dedup, counts = np.unique(subsample_read_index, return_counts=True)
        n_count_once = np.sum(counts == 1)
        read_total = frac_n_read
        saturation = round((1 - n_count_once / read_total) * 100, 2)
        df_cell_subsample = df_cell.loc[new_df.loc[index_dedup,]['oindex']]
        geneNum_median = float(df_cell_subsample.groupby(
        'Barcode').agg({'geneID': 'nunique'}).median())

        return saturation, geneNum_median

    @utils.add_log
    def downsample(self, df_cell):
        """saturation and median gene
        Returns:
            downsample dict             
                READ_FRACTION: float
                SATURATION: float
                MEDIAN_GENE_NUMBER: float
        """
        oindexList = []
        dcountList = []

        for row in df_cell.itertuples():
            duplicates = row.duplicate.split(',')
            ds = [int(d) for d in duplicates]
            for d in ds:
                oindexList.append(row.Index)
                dcountList.append(d)

        new_df = pd.DataFrame({'oindex':oindexList, 'dcount':dcountList})

        cell_read_index = np.array(new_df.index.repeat(new_df['dcount']), dtype='int32')
        np.random.shuffle(cell_read_index)

        downsample_dict = {
            READ_FRACTION: [0],
            SATURATION: [0],
            MEDIAN_GENE_NUMBER: [0],
        }

        for fraction in np.arange(0.1, 1.1, 0.1):
            saturation, geneNum_median = Count.sub_sample(fraction, df_cell, new_df, cell_read_index)
            fraction = round(fraction,1)
            downsample_dict[READ_FRACTION].append(fraction)
            downsample_dict[SATURATION].append(saturation)
            downsample_dict[MEDIAN_GENE_NUMBER].append(geneNum_median)

        df_downsample = pd.DataFrame(downsample_dict, columns=[READ_FRACTION, MEDIAN_GENE_NUMBER, SATURATION,])
        df_downsample.to_csv(self.downsample_file, index=False, sep='\t')
        return downsample_dict


@utils.add_log
def count(args):
    with Count(args, display_title="Cells") as runner:
        runner.run()


def get_opts_count(parser, sub_program):
    parser.add_argument('--genomeDir', help='Required. Genome directory.')
    parser.add_argument('--expected_cell_num', help='Default `3000`. Expected cell number.', default=3000)
    parser.add_argument(
        '--cell_calling_method',
        help=HELP_DICT['cell_calling_method'],
        choices=['auto', 'EmptyDrops_CR'],
        default='EmptyDrops_CR',
    )
    if sub_program:
        parser = s_common(parser)
        parser.add_argument('--count_detail', help='Required. File from featureCounts.', required=True)
        parser.add_argument(
            '--force_cell_num',
            help='Default `None`. Force the cell number to be this number. ',
        )
